use std::collections::HashMap;
use std::collections::hash_map::Entry;
use std::time::SystemTime;
use std::borrow::Borrow;
use std::sync::Mutex;

use sequoia_openpgp as openpgp;
use openpgp::cert::prelude::*;
use openpgp::Fingerprint;
use openpgp::KeyID;
use openpgp::KeyHandle;
use openpgp::policy::Policy;
use openpgp::Result;

use crate::Depth;
use crate::store::CertObject;
use crate::CertificationSet;
use crate::CertSynopsis;
use crate::store::Backend;
use crate::store::Store;
use crate::store::StoreError;

const TRACE: bool = false;

/// Manages a slice of `&Cert`.
///
/// `CertSlice` implements [`Backend`] and [`Store`] and thus can be
/// used as a certificate store by a [`Network`].
///
/// [`Network`]: crate::Network
pub struct CertSlice<'a> {
    // Indexed by KeyID.
    certs: HashMap<KeyID, Vec<ValidCert<'a>>>,

    // Certifications on a certificate.
    //
    // Example:
    //
    // ```
    // 0xA certifies <Carol, 0xC>.    \ CertificationSet
    // 0xA certifies <Caroline, 0xC>. /
    // 0xB certifies <Carol, 0xC>.    > CertificationSet
    // ```
    //
    // The entry for 0xC has two `CertificationSet`s: one for those
    // made by 0xA and one for those made by 0xB.
    //
    // This is a cache, which is derived from the certs, and is update
    // lazily.
    redge_cache: Mutex<HashMap<Fingerprint, Vec<CertificationSet>>>,

    // The reference time.  This is needed to compute the
    // certification sets.
    reference_time: SystemTime,
}

impl<'a> CertSlice<'a> {
    /// Returns a new `CertSlice`.
    ///
    /// The certificates must be deduped.  If there are multiple
    /// instances of a single certificate (i.e., two certificates with
    /// the same fingerprint), they WILL NOT be merged; the first one
    /// will be used, and additional ones will be silently ignored.
    pub fn new<T>(certs: &'a [Cert], policy: &'a dyn Policy, t: T)
        -> Result<Self>
    where T: Into<Option<SystemTime>>,
    {
        tracer!(TRACE, "CertSlice::new");

        let t = t.into().unwrap_or_else(SystemTime::now);

        let mut map: HashMap<KeyID, Vec<ValidCert>> = HashMap::new();

        'cert: for cert in certs.into_iter() {
            let vc = match cert.with_policy(policy, t) {
                Ok(vc) => vc,
                Err(err) => {
                    t!("Ignoring cert {}, which is not valid \
                        according to the policy: {}",
                       cert.fingerprint(), err);
                    continue;
                }
            };

            match map.entry(vc.keyid()) {
                Entry::Occupied(mut oe) => {
                    let v = oe.get_mut();
                    for c in v.iter_mut() {
                        if c.fingerprint() == vc.fingerprint() {
                            // We currently silently ignore
                            // duplicates.
                            t!("{} appears multiple times",
                               c.fingerprint());
                            continue 'cert;
                        }
                    }
                    v.push(vc);
                }
                Entry::Vacant(ve) => {
                    ve.insert(vec![ vc ]);
                }
            }
        }

        Ok(CertSlice {
            certs: map,
            redge_cache: Mutex::new(HashMap::new()),
            reference_time: t,
        })
    }
}

impl<'a> Backend for CertSlice<'a> {
    fn cert_object(&self, kh: &KeyHandle) -> Result<Vec<CertObject<'a>>>
    {
        tracer!(TRACE, "CertSlice::cert_object(backend)");

        let keyid = KeyID::from(kh);
        t!("{}", kh);

        let certs = self.certs.get(&keyid)
            .ok_or(StoreError::NotFound(kh.clone()))?;
        let certs: Vec<ValidCert> = match &kh {
            KeyHandle::Fingerprint(fpr) => {
                certs
                    .into_iter()
                    .find(|c| &c.fingerprint() == fpr)
                    .into_iter()
                    .cloned()
                    .collect()
            }
            KeyHandle::KeyID(_) => {
                // Return them all.
                certs
                    .into_iter()
                    .map(Into::into)
                    .cloned()
                    .collect()
            }
        };

        if certs.is_empty() {
            Err(StoreError::NotFound(kh.clone()).into())
        } else {
            Ok(certs.into_iter().map(|c| CertObject::ValidCert(c)).collect())
        }
    }

    fn list_cert_objects<'b>(&'b self)
        -> Box<dyn Iterator<Item=Fingerprint> + 'b>
    {
        tracer!(TRACE, "CertSlice::list_cert_objects(backend)");
        t!("");
        Box::new(self.certs.values().flat_map(|v| {
            v.into_iter().map(|c| c.fingerprint())
        }))
    }

    /// Prefills the cache.
    ///
    /// Prefilling the cache makes sense when you plan to examine most
    /// nodes and edges in the network.  It doesn't make sense if you
    /// are just authenticating a single or a few bindings.
    ///
    /// This function is multi-threaded.
    ///
    /// Errors are silently ignored and are propagated when the
    /// operation in question is executed directly.
    fn precompute(&self) {
        tracer!(TRACE, "CertSlice::precompute");

        use crossbeam::thread;
        use crossbeam::channel::unbounded as channel;

        // Avoid an extra level of indentation.
        let result = thread::scope(|thread_scope| {
        let certs: Vec<Fingerprint> = self.list().collect();

        // The threads.  We start them on demand.
        let threads = if certs.len() < 16 {
            // The keyring is small, limit the number of threads.
            2
        } else {
            // Use at least one and not more than we have cores.
            num_cpus::get().max(1)
        };

        // A communication channel for sending work to the workers.
        let (work_tx, work_rx) = channel();

        let mut threads_extant = Vec::new();

        for fpr in certs.into_iter() {
            if threads_extant.len() < threads {
                let tid = threads_extant.len();
                t!("Starting thread {} of {}",
                   tid, threads);

                let mut work = Some(Ok(fpr));

                // The thread's state.
                let work_rx = work_rx.clone();

                threads_extant.push(thread_scope.spawn(move |_| {
                    let mut results: Vec<(Fingerprint, Vec<CertificationSet>)>
                        = Vec::new();

                    loop {
                        match work.take().unwrap_or_else(|| work_rx.recv()) {
                            Err(_) => break,
                            Ok(fpr) => {
                                t!("Thread {} dequeuing {}!", tid, fpr);
                                // Silently ignore errors.  This will
                                // be caught later when the caller
                                // looks this one up.
                                if let Ok(redges)
                                    = self.certifications_of_uncached(&fpr)
                                {
                                    results.push((fpr, redges));
                                }
                            }
                        }
                    }

                    t!("Thread {} exiting", tid);

                    results
                }));
            } else {
                work_tx.send(fpr).unwrap();
            }
        }

        // When the threads see this drop, they will exit.
        drop(work_tx);

        let redges = threads_extant.into_iter().flat_map(|t| {
            let redges: Vec<(Fingerprint, Vec<CertificationSet>)>
                = t.join().unwrap();
            redges
        });

        let redges = HashMap::from_iter(redges);
        *self.redge_cache.lock().unwrap() = redges;

        }); // thread scope.

        // We're just caching results so we can ignore errors.
        if let Err(err) = result {
            t!("{:?}", err);
        }
    }
}

impl<'a> CertSlice<'a> {
    fn certifications_of_uncached<F>(&self, target: F)
        -> Result<Vec<CertificationSet>>
    where F: Borrow<Fingerprint>
    {
        let target = target.borrow();
        let target_keyid = KeyID::from(&KeyHandle::from(target));

        let cert = self.certs.get(&target_keyid)
            .ok_or(StoreError::NotFound(KeyHandle::from(target)))?
            // We searched by KeyID.  Prune any collisions.
            .into_iter()
            .find(|c| &c.fingerprint() == target)
            .into_iter()
            .next()
            .ok_or(StoreError::NotFound(KeyHandle::from(target)))?
            .clone();

        let redges = self.redges(cert, 0.into());

        Ok(redges)
    }
}

impl<'a> Store for CertSlice<'a> {
    fn reference_time(&self) -> SystemTime {
        self.reference_time
    }

    fn list<'b>(&'b self) -> Box<dyn Iterator<Item=Fingerprint> + 'b> {
        tracer!(TRACE, "CertSlice::list(store)");
        t!("");
        Box::new(self.certs.values().flat_map(|v| {
            v.into_iter().map(|c| c.fingerprint())
        }))
    }

    fn cert(&self, kh: &KeyHandle) -> Result<Vec<CertSynopsis>>
    {
        tracer!(TRACE, "CertSlice::cert(store)");

        t!("{}", kh);

        self.certs.get(&KeyID::from(kh))
            .map(|v| v.into_iter().map(|vc| vc.into()).collect())
            .ok_or(StoreError::NotFound(kh.clone()).into())
    }

    fn certifications_of(&self, target: &Fingerprint, _min_depth: Depth)
        -> Result<Vec<CertificationSet>>
    {
        tracer!(TRACE, "CertSlice::certifications_of(store)");

        t!("{}", target);

        let redge_cache = self.redge_cache.lock().unwrap();
        if let Some(redges) = redge_cache.get(target) {
            t!("Cache hit!");
            return Ok(redges.clone());
        }
        drop(redge_cache);

        t!("Cache miss!");

        let redges = self.certifications_of_uncached(target)?;

        self.redge_cache
            .lock().unwrap()
            .insert(target.clone(), redges.clone());

        Ok(redges)
    }
}
