use std::time::SystemTime;
use std::borrow::Borrow;
use std::borrow::Cow;
use std::cell::RefCell;
use std::collections::HashMap;
use std::sync::Mutex;

use sequoia_openpgp as openpgp;
use openpgp::Cert;
use openpgp::Fingerprint;
use openpgp::KeyHandle;
use openpgp::KeyID;
use openpgp::Result;
use openpgp::packet::UserID;
use openpgp::parse::Parse;
use openpgp::policy::NullPolicy;
use openpgp::policy::Policy;
use openpgp::serialize::Serialize;

use sequoia_net as net;

use openpgp_cert_d::CertD;

use crate::store::CertObject;
use crate::CertificationSet;
use crate::CertSynopsis;
use crate::store::Backend;
use crate::store::Store;
use crate::store::StoreError;
use crate::Depth;

const TRACE: bool = false;

const NP: NullPolicy = NullPolicy::new();

/// Attempts to fetch certificates from a keyserver.
pub struct KeyServer<'a, B>
    where B: Backend
{
    policy: &'a dyn Policy,
    reference_time: SystemTime,

    backend: B,
    keyserver: RefCell<net::KeyServer>,

    // Persistent disk cache.
    certd: Option<CertD>,
    // A certd currently only efficiently maps from Fingerprints to
    // Certificates.  At start up, we read in all the fingerprints,
    // which is equivalent to `find . -name` and thus relatively
    // cheap, and create a primary KeyID to primary Fingerprint map.
    // This is quite useful as lots of certifications only use the
    // Issuer subpacket, and not the Issuer Fingerprint subpacket.
    certd_by_keyid: RefCell<HashMap<KeyID, Fingerprint>>,

    // In-memory cache.
    cert_cache: Mutex<HashMap<KeyID, Vec<(Fingerprint, Option<Cert>)>>>,
    redge_all_cache: Mutex<HashMap<Fingerprint, Vec<CertificationSet>>>,
    redge_introducer_cache: Mutex<HashMap<Fingerprint, Vec<CertificationSet>>>,

    // Otherwise 'a is considered unused.
    phantom: std::marker::PhantomData<&'a B>,
}

impl<'a, B> KeyServer<'a, B>
    where B: Backend
{
    /// Returns a new `KeyServer`.
    ///
    /// When looking up certificates by Key ID or fingerprint, this
    /// first looks up the certificate in the underlying backend.  If
    /// no certificate is found, then it looks up the certificate on
    /// the configured key server.
    ///
    /// When looking up certificates by User ID or by email address
    /// (using [`Store::with_userid`] or [`Store::with_email`]), this
    /// simultaneously checks the corresponding WKD (if any), the
    /// configured key server, and the underlying backend.  The
    /// results are merged and cached.
    pub fn new(keyserver: net::KeyServer, backend: B,
               policy: &'a dyn Policy, reference_time: SystemTime)
        -> Result<Self>
    {
        tracer!(TRACE, "KeyServer::new");

        let (certd, certd_by_keyid) = match CertD::new() {
            Ok(certd) => {
                let by_keyid = if let Ok(iter) = certd.iter_fingerprints() {
                    HashMap::from_iter(
                        iter.filter_map(|fpr| {
                            let fpr = fpr.parse().ok()?;
                            Some((KeyID::from(&fpr), fpr))
                        }))
                } else {
                    HashMap::new()
                };

                (Some(certd), by_keyid)
            },
            Err(err) => {
                t!("Failed to open certd: {}", err);
                (None, HashMap::new())
            },
        };

        Ok(KeyServer {
            policy,
            reference_time,
            backend,
            keyserver: RefCell::new(keyserver),
            certd,
            certd_by_keyid: RefCell::new(certd_by_keyid),
            cert_cache: Mutex::new(HashMap::new()),
            redge_all_cache: Mutex::new(HashMap::new()),
            redge_introducer_cache: Mutex::new(HashMap::new()),
            phantom: std::marker::PhantomData,
        })
    }
}

impl<'a, B> Backend for KeyServer<'a, B>
    where B: Backend
{
    fn cert_object<'b>(&'b self, kh: &KeyHandle) -> Result<Vec<CertObject<'b>>>
    {
        tracer!(TRACE, "KeyServer::cert_object(backend)");
        t!("{}", kh);

        // Check our in-memory cache.
        let cert_cache = self.cert_cache.lock().unwrap();
        if let Some(certs) = cert_cache.get(&KeyID::from(kh)) {
            let certs = certs.into_iter().filter(|(fpr, _cert)| {
                kh.aliases(KeyHandle::from(fpr))
            }).collect::<Vec<_>>();

            if ! certs.is_empty() {
                t!("Cache hit!");
                return Ok(
                    certs
                        .into_iter()
                        .filter_map(|(_fpr, cert)| {
                            if let Some(cert) = cert {
                                Some(CertObject::Cert(cert.clone()))
                            } else {
                                None
                            }
                        })
                        .collect());
            }
        }
        drop(cert_cache);

        // Check the underlying backend.
        if let Ok(r) = self.backend.cert_object(&kh) {
            return Ok(r);
        }

        // Check the persistent cache.
        if let Some(certd) = self.certd.as_ref() {
            let fpr = match kh {
                KeyHandle::Fingerprint(ref fpr) => {
                    Some(fpr.clone())
                }
                KeyHandle::KeyID(ref keyid) => {
                    self.certd_by_keyid.borrow().get(keyid).map(Clone::clone)
                }
            };

            if let Some(fpr) = fpr {
                match certd.get(&fpr.to_string()) {
                    Ok(Some((_tag, data))) => {
                        t!("  Found {} in disk cache.", kh);
                        match Cert::from_bytes(&data) {
                            Ok(cert) => {
                                if cert.fingerprint() == fpr {
                                    self.cert_cache.lock().unwrap().insert(
                                        KeyID::from(kh),
                                        vec![ (fpr.clone(), Some(cert.clone())) ]);
                                    return Ok(vec![ CertObject::Cert(cert) ]);
                                } else {
                                    t!("cert-d returned wrong certificate: \
                                        got {}, looked up: {}",
                                       cert.fingerprint(), fpr);
                                }
                            }
                            Err(err) => t!("Parsing data returned by cert-d: {}",
                                           err),
                        }
                    }
                    // Not found.
                    Ok(None) => (),
                    Err(err) => {
                        t!("  Reading {} from cert-d: {}", kh, err);
                    }
                }
            }
        }

        // Now try the key server.
        eprint!("Looking {} up on the keyserver... ",
                KeyID::from(kh));

        let rt = tokio::runtime::Runtime::new().unwrap();
        let r = rt.block_on(async {
            self.keyserver.borrow_mut().get(kh.clone()).await
        });
        match r {
            Ok(cert) => {
                let possible_userid =
                    cert.with_policy(self.policy, self.reference_time)
                        .or_else(|_| cert.with_policy(&NP, self.reference_time))
                        .and_then(|vc| vc.primary_userid())
                        .map(|ua| {
                            String::from_utf8_lossy(
                                ua.userid().value())
                        })
                        .unwrap_or_else(|_| {
                            cert.userids().next()
                                .map(|ua| {
                                    String::from_utf8_lossy(
                                        ua.userid().value())
                                })
                                .unwrap_or(Cow::Borrowed("<unknown>"))
                        });

                if ! cert.key_handle().aliases(kh) {
                    eprintln!("keyserver returned the wrong key: \
                               {} ({:?}), wanted: {}",
                              cert.key_handle(), possible_userid, kh);
                    Err(StoreError::NotFound(
                        KeyHandle::from(kh.clone())).into())
                } else if let Err(err) = cert.with_policy(
                    self.policy, self.reference_time)
                {
                    eprintln!("invalid ({:?}): {}.",
                              possible_userid, err);
                    self.cert_cache.lock().unwrap().insert(
                        cert.keyid(),
                        vec![ (cert.fingerprint(), None) ]);
                    Err(StoreError::NotFound(
                        KeyHandle::from(kh.clone())).into())
                } else {
                    eprintln!("ok ({:?}, {} certifications).",
                              possible_userid,
                              cert.userids()
                                  .map(|u| u.certifications().count())
                                  .sum::<usize>());

                    if let Err(err) = self.cache_cert(&cert) {
                        t!("Caching {}: {}", kh, err);
                    }
                    Ok(vec![ CertObject::Cert(cert) ])
                }
            }
            Err(err) => {
                eprintln!("error: {}", err);
                // Only cache misses for look ups when we know the
                // fingerprint.
                if let KeyHandle::Fingerprint(fpr) = kh {
                    self.cert_cache.lock().unwrap().insert(
                        KeyID::from(fpr),
                        vec![ (fpr.clone(), None) ]);
                }
                Err(err)
            }
        }
    }

    fn list_cert_objects<'b>(&'b self)
        -> Box<dyn Iterator<Item=Fingerprint> + 'b>
    {
        tracer!(TRACE, "KeyServer::list_cert_objects(backend)");
        t!("");

        // The backend, plus anything in our cache.
        let mut certs = self.backend.list_cert_objects()
            .chain(self.cert_cache.lock().unwrap().values()
                   .flat_map(|certs| {
                       certs.iter().filter_map(|(fpr, cert)| {
                           if cert.is_some() {
                               Some(fpr.clone())
                           } else {
                               None
                           }
                       })
                   }))
            .collect::<Vec<Fingerprint>>();
        certs.sort();
        certs.dedup();

        Box::new(certs.into_iter())
    }

    fn precompute(&self) {
        self.backend.precompute()
    }
}

impl<'a, B> KeyServer<'a, B>
    where B: Backend
{
    // Caches the certificate in the in-memory and in the on-disk
    // caches.
    fn cache_cert(&self, cert: &Cert) -> Result<()> {
        tracer!(TRACE, "KeyServer::cache_cert");

        // Insert in the in-memory cache.
        self.cert_cache.lock().unwrap().insert(
            cert.keyid(),
            vec![ (cert.fingerprint(), Some(cert.clone())) ]);

        // Insert in the disk cache.
        let certd = if let Some(certd) = self.certd.as_ref() {
            certd
        } else {
            // Caching on disk is disabled, we're done.
            t!("Disk cache disabled.");
            return Ok(());
        };

        // Note: if we got secrets, we discard them.  That's the right
        // choice.
        let mut cert_bytes = Vec::new();
        cert.serialize(&mut cert_bytes)?;

        certd.try_insert(
            cert_bytes.into_boxed_slice(),
            |cert_bytes, disk_bytes| {
                let disk_bytes = if let Some(disk_bytes) = disk_bytes {
                    disk_bytes
                } else {
                    // Nothing to merge.
                    return Ok(cert_bytes);
                };

                match Cert::from_bytes(&disk_bytes) {
                    Ok(disk) => {
                        if cert.fingerprint() != disk.fingerprint() {
                            // cert-d returned data with the wrong
                            // fingerprint.  Don't try to fix the cert-d,
                            // just silently fail.  (It doesn't really
                            // matter, because this is just a cache.)
                            t!("certd returned wrong cert, \
                                got: {}, expected: {}",
                               disk.fingerprint(), cert.fingerprint());
                            return Ok(disk_bytes);
                        }

                        // If the on-disk version has secrets, we
                        // preserve them.
                        Ok(disk.merge_public_and_secret(cert.clone())
                           .map(|merged| {
                               let mut bytes = Vec::new();
                               merged.as_tsk()
                                   .serialize(&mut bytes)
                                   .expect("serializing to a vec");
                               bytes.into_boxed_slice()
                           })?)
                    }
                    Err(err) => {
                        // We failed to parse disk_bytes.
                        t!("Failed to parse data returned by certd: {}",
                           err);
                        Ok(cert_bytes)
                    }
                }
            })?;

        self.certd_by_keyid.borrow_mut()
            .insert(cert.keyid(), cert.fingerprint());

        Ok(())
    }

    fn certifications_of_uncached<F>(&self, target: F, min_depth: Depth)
        -> Result<Vec<CertificationSet>>
    where F: Borrow<Fingerprint>
    {
        let target = target.borrow();

        let cert = self.cert_object_by_fpr(target)?;

        // Turn invalid certificate errors into NotFound errors.
        let vc = cert.with_policy(self.policy, self.reference_time)
            .map_err(|_| StoreError::NotFound(KeyHandle::from(target)))?;

        let redges = self.redges(vc, min_depth.min(Depth::from(1)));

        Ok(redges)
    }
}

impl<'a, B> KeyServer<'a, B>
    where B: Backend + Store
{
    // Search for certificates by User ID or email address.  Note: the
    // keyserver does a substring search so this works for email
    // addresses by default.
    fn search(&self, userid: &UserID, email: bool) -> Vec<Cert> {
        tracer!(TRACE, "KeyServer::search");

        let id = String::from_utf8_lossy(userid.value());
        t!("{}", id);
        eprint!("Looking {:?} up on the keyserver... ", id);

        let email = if email {
            if let Ok(Some(email)) = userid.email() {
                Some(email)
            } else {
                eprintln!("invalid email address");
                return Vec::new();
            }
        } else {
            None
        };

        let rt = tokio::runtime::Runtime::new().unwrap();
        let (ks, wkd) = rt.block_on(async {
            // Query the keyserver.
            let mut ks = self.keyserver.borrow_mut();
            let ks = ks.search(userid.clone());

            // And the WKD (if it appears to be an email address).
            let wkd = async {
                if let Some(email) = email.as_ref() {
                    net::wkd::get(email).await
                } else {
                    // If it is not an email, it's not an error.
                    Ok(Vec::new())
                }
            };

            tokio::join!(ks, wkd)
        });

        let mut certs = Vec::new();
        match ks {
            Ok(c) => {
                t!("Key server returned {} results", c.len());
                certs.extend(c);
            },
            Err(err) => t!("Key server response: {}", err),
        }
        match wkd {
            Ok(c) => {
                t!("WKD server returned {} results", c.len());
                certs.extend(c);
            },
            Err(err) => t!("WKD server response: {}", err),
        }

        // Sort, merge, and cache.
        certs.sort_by_key(|c| c.fingerprint());
        certs.dedup_by(|a, b| {
            if a.fingerprint() != b.fingerprint() {
                return false;
            }

            // b is kept.  So merge into b.
            match b.clone().merge_public(a.clone()) {
                Ok(combined) => *b = combined,
                Err(err) => {
                    t!("Merging copies of {}: {}",
                       a.keyid(), err);
                }
            }

            true
        });

        eprintln!(
            "Got {} results:\n  {}",
           certs.len(),
           certs.iter().map(|cert| {
               format!(
                   "{} ({})",
                   cert.keyid().to_hex(),
                   cert.with_policy(self.policy, self.reference_time)
                       .and_then(|vc| vc.primary_userid())
                       .map(|ua| {
                           String::from_utf8_lossy(ua.userid().value())
                       })
                       .unwrap_or_else(|_| {
                           cert.userids().next()
                               .map(|ua| {
                                   String::from_utf8_lossy(ua.userid().value())
                               })
                               .unwrap_or(Cow::Borrowed("<unknown>"))
                       }))
           })
           .collect::<Vec<_>>()
           .join("\n  "));

        // Merge in the backend versions.
        let mut certs = certs.into_iter()
            .map(|cert| {
                let fpr = cert.fingerprint();
                if let Ok(copy) = self.backend.cert_object_by_fpr(&fpr) {
                    cert.merge_public(copy.cert().clone()).expect("same cert")
                } else {
                    cert
                }
            })
            .collect::<Vec<Cert>>();

        // Cache the results.
        if ! certs.is_empty() {
            let mut cert_cache = self.cert_cache.lock().unwrap();

            for cert in certs.iter() {
                let fpr = cert.fingerprint();
                let keyid = KeyID::from(&fpr);

                let mut inserted = false;
                if let Some(cache_entries) = cert_cache.get_mut(&keyid) {
                    for mut entry in cache_entries.iter_mut() {
                        if entry.0 == fpr {
                            entry.1 = Some(cert.clone());
                            inserted = true;
                            break;
                        }
                    }
                }

                if ! inserted {
                    cert_cache.insert(
                        keyid, vec![ (fpr, Some(cert.clone())) ]);
                }
            }
        }

        // Add the results from the backend.
        let backend_fprs: Vec<Fingerprint> = if let Some(email) = email {
            self.backend.with_email(&email)
                .into_iter()
                .map(|(fpr, _)| fpr)
                .collect::<Vec<_>>()
        } else {
            self.backend.with_userid(userid.clone())
        };

        assert!(certs
                .windows(2)
                .all(|certs| {
                    certs[0].fingerprint() < certs[1].fingerprint()
                }),
                "not sorted");

        let mut backend_certs = Vec::new();
        for fpr in backend_fprs.into_iter() {
            if let Ok(_i) = certs.binary_search_by(|c| {
                c.fingerprint().cmp(&fpr)
            })
            {
                // We already have it (and we already merged it in the
                // backend copy above).  We're done.
            } else {
                // Add it to our results.
                match self.cert_object_by_fpr(&fpr) {
                    Ok(c) => backend_certs.push(c.cert().clone()),
                    Err(err) =>
                        t!("with_userid returned {}, but lookup failed: {}",
                           fpr, err),
                }
            }
        }

        certs.extend(backend_certs.into_iter());

        certs
    }
}

impl<'a, B> Store for KeyServer<'a, B>
    where B: Backend + Store
{
    fn reference_time(&self) -> SystemTime {
        self.backend.reference_time()
    }

    fn list<'b>(&'b self) -> Box<dyn Iterator<Item=Fingerprint> + 'b> {
        tracer!(TRACE, "KeyServer::list(store)");
        t!("");

        self.list_cert_objects()
    }

    fn cert(&self, kh: &KeyHandle) -> Result<Vec<CertSynopsis>> {
        tracer!(TRACE, "KeyServer::cert(store)");
        t!("{}", kh);

        let certs = self.cert_object(kh)?
            .into_iter()
            .filter_map(|cert| {
                match cert.with_policy(self.policy, self.reference_time) {
                    Ok(vc) => Some(CertSynopsis::from(vc)),
                    Err(err) => {
                        t!("{}: {}", cert.keyid(), err);
                        None
                    }
                }
            })
            .collect::<Vec<CertSynopsis>>();

        if certs.is_empty() {
            Err(StoreError::NotFound(kh.clone()).into())
        } else {
            Ok(certs)
        }
    }

    fn certifications_of(&self, target: &Fingerprint, min_depth: Depth)
        -> Result<Vec<CertificationSet>>
    {
        tracer!(TRACE, "KeyServer::certifications_of(store)");

        t!("{}", target);

        let redge_all_cache = self.redge_all_cache.lock().unwrap();
        if let Some(redges) = redge_all_cache.get(target) {
            t!("Cache hit!");
            return Ok(redges.clone());
        }
        drop(redge_all_cache);

        if min_depth > 0.into() {
            let redge_introducer_cache
                = self.redge_introducer_cache.lock().unwrap();
            if let Some(redges) = redge_introducer_cache.get(target) {
                t!("Cache hit!");
                return Ok(redges.clone());
            }
            drop(redge_introducer_cache);
        }

        t!("Cache miss!");

        let redges = self.certifications_of_uncached(
            target, min_depth.min(Depth::from(1)))?;

        if min_depth == 0.into() {
            self.redge_all_cache
                .lock().unwrap()
                .insert(target.clone(), redges.clone());
        } else {
            self.redge_introducer_cache
                .lock().unwrap()
                .insert(target.clone(), redges.clone());
        }

        Ok(redges)
    }

    fn with_userid(&self, userid: UserID) -> Vec<Fingerprint> {
        tracer!(TRACE, "KeyServer::with_userid");
        t!("Looking up {}", String::from_utf8_lossy(userid.value()));

        let mut fprs = self.backend.with_userid(userid.clone());

        let certs = self.search(&userid, false);

        // Only return the certificates that actually match.
        fprs.extend(certs.into_iter().flat_map(|c| {
            c.userids()
                .filter_map(|ua| {
                    if ua.userid() == &userid {
                        Some(c.fingerprint())
                    } else {
                        None
                    }
                })
                .collect::<Vec<_>>()
                .into_iter()
        }));

        fprs
    }

    fn with_email(&self, email: &str) -> Vec<(Fingerprint, UserID)> {
        // We don't cache the returned results, as we prefer the
        // versions in the local store.

        tracer!(TRACE, "KeyServer::with_email");
        t!("Looking up {}", email);

        let userid = UserID::from(format!("<{}>", email));
        if let Ok(Some(email_check)) = userid.email() {
            if email != email_check {
                // Does not appear to be an email address.
                return Vec::new();
            }
        } else {
            // Does not appear to be an email address.
            return Vec::new();
        }

        let email_normalized = if let Ok(Some(e)) = userid.email_normalized() {
            e
        } else {
            // Does not appear to be an email address.
            return Vec::new();
        };

        let mut fprs = self.backend.with_email(email);

        let certs = self.search(&userid, true);

        // Only return the certificates that actually match.
        fprs.extend(certs.into_iter().flat_map(|c| {
            c.userids()
                .filter_map(|ua| {
                    if let Ok(Some(e)) = ua.email_normalized() {
                        if e == email_normalized {
                            return Some((c.fingerprint(), ua.userid().clone()));
                        }
                    }
                    None
                })
                .collect::<Vec<_>>()
                .into_iter()
        }));

        fprs
    }
}

