use std::collections::HashMap;
use std::collections::hash_map::Entry;
use std::time::SystemTime;
use std::borrow::Borrow;
use std::sync::Mutex;

use sequoia_openpgp as openpgp;
use openpgp::cert::prelude::*;
use openpgp::cert::raw::RawCert;
use openpgp::cert::raw::RawCertParser;
use openpgp::Fingerprint;
use openpgp::KeyID;
use openpgp::KeyHandle;
use openpgp::parse::Parse;
use openpgp::policy::Policy;
use openpgp::Result;

use crate::Depth;
use crate::store::CertObject;
use crate::CertificationSet;
use crate::CertSynopsis;
use crate::store::Backend;
use crate::store::Store;
use crate::store::StoreError;

const TRACE: bool = false;

struct LazyCert<'a> {
    // The certificate's bytes.
    raw: RawCert<'a>,

    // The parsed certificate.  This is only created on demand, hence
    // the use of the Mutex for interior mutability.  If an error
    // occurs parsing the certificate data, then None is stored.
    cert: Mutex<Option<Option<Cert>>>,
}

impl<'a> LazyCert<'a> {
    fn new(raw: RawCert<'a>) -> Self {
        Self {
            raw,
            cert: Mutex::new(None),
        }
    }

    fn fingerprint(&self) -> Fingerprint {
        self.raw.fingerprint()
    }

    // Return the parsed certificate.
    //
    // If the certificate hasn't been parsed yet, parses it and caches
    // it.  If an error occurs parsing the certificate, then None is
    // returned.
    fn parsed(&self) -> Option<Cert> {
        tracer!(TRACE, "LazyCert::parsed");

        let mut cert = self.cert.lock().unwrap();

        if cert.is_none() {
            match (&self.raw).try_into() {
                Ok(c) => *cert = Some(Some(c)),
                Err(err) => {
                    t!("Parsing certificate {}: {}",
                       self.fingerprint(), err);
                    *cert = Some(None);
                }
            }
        }
        (*cert).as_ref().expect("some").clone()
    }
}

/// Manages a slice of bytes or a slice of `&RawCert`.
///
/// `RawCerts` implements [`Backend`] and [`Store`] and thus can be
/// used as a certificate store by a [`Network`].
///
/// [`Network`]: crate::Network
pub struct RawCerts<'a>
{
    // Indexed by KeyID.
    certs: HashMap<KeyID, Vec<LazyCert<'a>>>,

    // Certifications on a certificate.
    //
    // Example:
    //
    // ```
    // 0xA certifies <Carol, 0xC>.    \ CertificationSet
    // 0xA certifies <Caroline, 0xC>. /
    // 0xB certifies <Carol, 0xC>.    > CertificationSet
    // ```
    //
    // The entry for 0xC has two `CertificationSet`s: one for those
    // made by 0xA and one for those made by 0xB.
    //
    // This is a cache, which is derived from the certs, and is update
    // lazily.
    redge_all_cache: Mutex<HashMap<Fingerprint, Vec<CertificationSet>>>,
    redge_introducer_cache: Mutex<HashMap<Fingerprint, Vec<CertificationSet>>>,

    // The policy.  This is needed to compute the certification sets.
    policy: &'a dyn Policy,

    // The reference time.  This is needed to compute the
    // certification sets.
    reference_time: SystemTime,
}

impl<'a> RawCerts<'a>
{
    /// Returns a new `RawCerts`.
    ///
    /// The certificates must be deduped.  If there are multiple
    /// instances of a single certificate (i.e., two certificates with
    /// the same fingerprint), they WILL NOT be merged; the first one
    /// will be used, and additional ones will be silently ignored.
    pub fn new<T>(bytes: &'a [u8], policy: &'a dyn Policy, t: T)
        -> Result<Self>
    where T: Into<Option<SystemTime>>,
    {
        tracer!(TRACE, "RawCerts::new");

        let certs: Vec<RawCert<'a>> = RawCertParser::from_bytes(bytes)?
            .filter_map(|c| {
                match c {
                    Ok(c) => Some(c),
                    Err(err) => {
                        t!("Parsing raw certificate: {}", err);
                        None
                    }
                }
            })
            .collect();
        Self::from_raw_certs(&certs, policy, t)
    }

    /// Returns a new `RawCerts`.
    ///
    /// The certificates must be deduped.  If there are multiple
    /// instances of a single certificate (i.e., two certificates with
    /// the same fingerprint), they WILL NOT be merged; the first one
    /// will be used, and additional ones will be silently ignored.
    pub fn from_raw_certs<T>(certs: &[RawCert<'a>],
                             policy: &'a dyn Policy, t: T)
        -> Result<Self>
    where T: Into<Option<SystemTime>>,
    {
        tracer!(TRACE, "RawCerts::from_raw_certs");

        let t = t.into().unwrap_or_else(SystemTime::now);

        let mut map: HashMap<KeyID, Vec<LazyCert<'a>>>
            = HashMap::new();

        'cert: for cert in certs.into_iter() {
            match map.entry(cert.keyid()) {
                Entry::Occupied(mut oe) => {
                    let v = oe.get_mut();
                    for c in v.iter_mut() {
                        if c.fingerprint() == cert.fingerprint() {
                            // We currently silently ignore
                            // duplicates.
                            t!("{} appears multiple times",
                               c.fingerprint());
                            continue 'cert;
                        }
                    }
                    v.push(LazyCert::new(cert.clone()));
                }
                Entry::Vacant(ve) => {
                    ve.insert(vec![ LazyCert::new(cert.clone()) ]);
                }
            }
        }

        Ok(RawCerts {
            certs: map,
            redge_all_cache: Mutex::new(HashMap::new()),
            redge_introducer_cache: Mutex::new(HashMap::new()),
            policy: policy,
            reference_time: t,
        })
    }
}

impl<'a> Backend for RawCerts<'a>
{
    fn cert_object(&self, kh: &KeyHandle) -> Result<Vec<CertObject<'a>>>
    {
        tracer!(TRACE, "RawCerts::cert_object(backend)");

        let keyid = KeyID::from(kh);
        t!("{}", kh);

        let certs = self.certs.get(&keyid)
            .ok_or(StoreError::NotFound(kh.clone()))?;
        let certs: Vec<CertObject> = match &kh {
            KeyHandle::Fingerprint(fpr) => {
                certs
                    .into_iter()
                    .find(|c| &c.fingerprint() == fpr)
                    .into_iter()
                    .filter_map(|c| c.parsed())
                    .map(|c| CertObject::Cert(c))
                    .collect()
            }
            KeyHandle::KeyID(_) => {
                // Return them all.
                certs
                    .into_iter()
                    .filter_map(|c| c.parsed())
                    .map(|c| CertObject::Cert(c))
                    .collect()
            }
        };

        if certs.is_empty() {
            Err(StoreError::NotFound(kh.clone()).into())
        } else {
            Ok(certs)
        }
    }

    fn list_cert_objects<'b>(&'b self)
        -> Box<dyn Iterator<Item=Fingerprint> + 'b>
    {
        tracer!(TRACE, "RawCerts::list_cert_objects(backend)");
        t!("");
        Box::new(self.certs.values().flat_map(|v| {
            v.into_iter().map(|c| c.fingerprint())
        }))
    }

    /// Prefills the cache.
    ///
    /// Prefilling the cache makes sense when you plan to examine most
    /// nodes and edges in the network.  It doesn't make sense if you
    /// are just authenticating a single or a few bindings.
    ///
    /// This function is multi-threaded.
    ///
    /// Errors are silently ignored and are propagated when the
    /// operation in question is executed directly.
    fn precompute(&self) {
        tracer!(TRACE, "RawCerts::precompute");

        use crossbeam::thread;
        use crossbeam::channel::unbounded as channel;

        // Avoid an extra level of indentation.
        let result = thread::scope(|thread_scope| {
        let mut certs: Vec<&LazyCert>
            = self.certs.values().flatten().collect();
        let cert_count = certs.len();

        // The threads.  We start them on demand.
        let threads = if cert_count < 16 {
            // The keyring is small, limit the number of threads.
            2
        } else {
            // Sort the certificates so they are ordered from most
            // packets to least.  More packets implies more work, and
            // this will hopefully result in a more equal distribution
            // of load.
            certs.sort_unstable_by_key(|c| usize::MAX - c.raw.count());

            // Use at least one and not more than we have cores.
            num_cpus::get().max(1)
        };

        // A communication channel for sending work to the workers.
        let (work_tx, work_rx) = channel();

        let mut threads_extant = Vec::new();

        for cert in certs.into_iter() {
            if threads_extant.len() < threads {
                let tid = threads_extant.len();
                t!("Starting thread {} of {}",
                   tid, threads);

                let mut work = Some(Ok(cert));

                // The thread's state.
                let work_rx = work_rx.clone();

                threads_extant.push(thread_scope.spawn(move |_| {
                    let mut results: Vec<(Fingerprint, Vec<CertificationSet>)>
                        = Vec::with_capacity(16 + cert_count / threads);

                    loop {
                        match work.take().unwrap_or_else(|| work_rx.recv()) {
                            Err(_) => break,
                            Ok(cert) => {
                                let fpr = cert.fingerprint();
                                t!("Thread {} dequeuing {}!", tid, fpr);
                                // Silently ignore errors.  This will
                                // be caught later when the caller
                                // looks this one up.

                                // Turn parse errors into NotFound errors.
                                if let Some(cert) = cert.parsed() {
                                    // Turn invalid certificate errors
                                    // into NotFound errors.
                                    if let Ok(vc) = cert.with_policy(
                                        self.policy, self.reference_time)
                                    {
                                        results.push((fpr, self.redges(vc, 0.into())));
                                    }
                                }
                            }
                        }
                    }

                    t!("Thread {} exiting", tid);

                    results
                }));
            } else {
                work_tx.send(cert).unwrap();
            }
        }

        // When the threads see this drop, they will exit.
        drop(work_tx);

        let redges = threads_extant.into_iter().flat_map(|t| {
            let redges: Vec<(Fingerprint, Vec<CertificationSet>)>
                = t.join().unwrap();
            redges
        });

        let redges = HashMap::from_iter(redges);
        *self.redge_all_cache.lock().unwrap() = redges;

        }); // thread scope.

        // We're just caching results so we can ignore errors.
        if let Err(err) = result {
            t!("{:?}", err);
        }
    }
}

impl<'a> RawCerts<'a>
{
    fn certifications_of_uncached<F>(&self, target: F, min_depth: Depth)
        -> Result<Vec<CertificationSet>>
    where F: Borrow<Fingerprint>
    {
        let target = target.borrow();
        let target_keyid = KeyID::from(&KeyHandle::from(target));

        let cert = self.certs.get(&target_keyid)
            .ok_or(StoreError::NotFound(KeyHandle::from(target)))?
            // We searched by KeyID.  Prune any collisions.
            .into_iter()
            .find(|c| &c.fingerprint() == target)
            .into_iter()
            .next()
            .ok_or(StoreError::NotFound(KeyHandle::from(target)))?;

        // Turn parse errors into NotFound errors.
        let cert = cert.parsed()
            .ok_or(StoreError::NotFound(KeyHandle::from(target)))?;
        // Turn invalid certificate errors into NotFound errors.
        let vc = cert.with_policy(self.policy, self.reference_time)
            .map_err(|_| StoreError::NotFound(KeyHandle::from(target)))?;

        let redges = self.redges(vc, min_depth);

        Ok(redges)
    }
}

impl<'a> Store for RawCerts<'a>
{
    fn reference_time(&self) -> SystemTime {
        self.reference_time
    }

    fn list<'b>(&'b self) -> Box<dyn Iterator<Item=Fingerprint> + 'b> {
        tracer!(TRACE, "RawCerts::list(store)");
        t!("");
        Box::new(self.certs.values().flat_map(|v| {
            v.into_iter().map(|c| c.fingerprint())
        }))
    }

    fn cert(&self, kh: &KeyHandle) -> Result<Vec<CertSynopsis>>
    {
        tracer!(TRACE, "RawCerts::cert(store)");

        t!("{}", kh);

        self.certs.get(&KeyID::from(kh))
            .map(|v| {
                v.into_iter()
                    .filter_map(|c| c.parsed())
                    .filter_map(|c| {
                        c.with_policy(self.policy, self.reference_time())
                            .map(Into::into)
                            .ok()
                    })
                    .collect()
            })
            .ok_or(StoreError::NotFound(kh.clone()).into())
    }

    fn certifications_of(&self, target: &Fingerprint, min_depth: Depth)
        -> Result<Vec<CertificationSet>>
    {
        tracer!(TRACE, "RawCerts::certifications_of(store)");

        t!("{}", target);

        let redge_all_cache = self.redge_all_cache.lock().unwrap();
        if let Some(redges) = redge_all_cache.get(target) {
            t!("Certification cache hit!");
            return Ok(redges.clone());
        }
        drop(redge_all_cache);

        if min_depth > 0.into() {
            let redge_introducer_cache = self.redge_introducer_cache.lock().unwrap();
            if let Some(redges) = redge_introducer_cache.get(target) {
                t!("Introducer cache hit!");
                return Ok(redges.clone());
            }
            drop(redge_introducer_cache);
        }

        t!("Cache miss!");

        let redges = self.certifications_of_uncached(
            target, min_depth.min(Depth::from(1)))?;

        if min_depth == 0.into() {
            self.redge_all_cache
                .lock().unwrap()
                .insert(target.clone(), redges.clone());
        } else {
            self.redge_introducer_cache
                .lock().unwrap()
                .insert(target.clone(), redges.clone());
        }

        Ok(redges)
    }
}
