#[cfg(test)]
mod integration {
    use std::path;
    use std::time;

    use assert_cmd::Command;
    use predicates::prelude::*;

    use sequoia_openpgp as openpgp;

    use openpgp::Fingerprint;
    use openpgp::Result;
    use openpgp::packet::UserID;

    fn dir() -> path::PathBuf {
        path::Path::new("tests").join("data")
    }

    fn test<'a, R, S>(keyring: &str, trust_root: R,
                      sqwot_args: &[&str], command: &str, args: &[&str],
                      amount: usize,
                      userid: Option<&UserID>,
                      target: Option<&Fingerprint>,
                      success: bool, output: &[(usize, S)]) -> Result<()>
    where R: Into<Option<&'a Fingerprint>>,
          S: AsRef<str>,
    {
        let trust_root = trust_root.into();

        let mut cmd = Command::cargo_bin("sq-wot")?;
        cmd.current_dir(&dir())
            // Use the default crypto policy.
            .env("SEQUOIA_CRYPTO_POLICY", "")
            .args(sqwot_args)
            .args(&["--keyring", keyring]);
        if let Some(trust_root) = trust_root {
            cmd.args(&["--trust-root", &trust_root.to_string()]);
        }
        cmd.arg(command);
        if let Some(target) = target {
            cmd.arg(&target.to_string());
        }
        if let Some(userid) = userid {
            cmd.arg(format!("{}", String::from_utf8_lossy(userid.value())));
        }
        for arg in args {
            cmd.arg(arg);
        }
        cmd.args(&[
            "--trust-amount", &format!("{}", amount),
        ]);

        let assertion = cmd.assert();

        if success {
            let assertion = assertion.success();

            let stdout
                = String::from_utf8_lossy(&assertion.get_output().stdout);
            for (expected_occurrences, s) in output {
                let s = s.as_ref();
                let occurrences = stdout.split(s).count() - 1;

                assert_eq!(occurrences, *expected_occurrences,
                           "Failed to find: '{}' {} times\n\
                            in stdout:\n\
                            {}",
                           s, *expected_occurrences, stdout);
            }
        } else {
            assertion.code(predicate::eq(1));
        }

        Ok(())
    }

    // Test authenticating a binding (User ID and certificate).
    #[test]
    #[allow(unused)]
    fn authenticate() -> Result<()> {
        let alice_fpr: Fingerprint =
            "85DAB65713B2D0ABFC5A4F28BC10C9CE4A699D8D"
           .parse().expect("valid fingerprint");
        let alice_uid
            = UserID::from("<alice@example.org>");

        let dave_fpr: Fingerprint =
            "329D5AAF73DC70B4E3DD2D11677CB70FFBFE1281"
           .parse().expect("valid fingerprint");
        let dave_uid
            = UserID::from("<dave@example.org>");

        let ellen_fpr: Fingerprint =
            "A7319A9B166AB530A5FBAC8AB43CA77F7C176AF4"
           .parse().expect("valid fingerprint");
        let ellen_uid
            = UserID::from("<ellen@example.org>");
        // Certified by: 329D5AAF73DC70B4E3DD2D11677CB70FFBFE1281

        let t = |amount: usize, userid: &UserID,
                 target: &Fingerprint, success: bool|
        {
            let output: &[(usize, &str)] = &[
                (1, &format!("[✓] {} {}: ", target, userid)),
                (1, "[✓] ")];
            test("simple.pgp", &alice_fpr, &[], "authenticate", &[],
                 amount, Some(userid), Some(target),
                 success, output)
        };

        t(100, &dave_uid, &dave_fpr, true)?;
        t(120, &dave_uid, &dave_fpr, false)?;

        // Not enough depth.
        t(100, &ellen_uid, &ellen_fpr, false)?;
        t(120, &ellen_uid, &ellen_fpr, false)?;

        // No such User ID on dave's key.
        t(100, &ellen_uid, &dave_fpr, false)?;
        t(120, &ellen_uid, &dave_fpr, false)?;

        Ok(())
    }

    // Test authenticating bindings where we match on just the email
    // address, not the whole User ID.
    #[test]
    #[allow(unused)]
    fn authenticate_email() -> Result<()> {
        let alice_fpr: Fingerprint =
            "85DAB65713B2D0ABFC5A4F28BC10C9CE4A699D8D"
           .parse().expect("valid fingerprint");
        let alice_uid
            = UserID::from("<alice@example.org>");

        let dave_fpr: Fingerprint =
            "329D5AAF73DC70B4E3DD2D11677CB70FFBFE1281"
           .parse().expect("valid fingerprint");
        let dave_uid
            = UserID::from("<dave@example.org>");
        let dave_email = UserID::from("dave@example.org");
        let dave_email_uc1 = UserID::from("DAVE@example.org");
        let dave_email_uc2 = UserID::from("DAVE@EXAMPLE.ORG");

        let ellen_fpr: Fingerprint =
            "A7319A9B166AB530A5FBAC8AB43CA77F7C176AF4"
           .parse().expect("valid fingerprint");
        let ellen_uid
            = UserID::from("<ellen@example.org>");
        let ellen_email = UserID::from("ellen@example.org");
        // Certified by: 329D5AAF73DC70B4E3DD2D11677CB70FFBFE1281

        let t = |amount: usize, pattern: &UserID,
                 userid: &UserID, target: &Fingerprint, success: bool|
        {
            let output: &[(usize, &str)] = &[
                (1, &format!("[✓] {} {}: ", target, userid)),
                (1, "[✓] ")];
            test("simple.pgp", &alice_fpr,
                 &[], "authenticate", &["--email"],
                 amount, Some(pattern), Some(target),
                 success, output)
        };

        t(100, &dave_uid, &dave_uid, &dave_fpr, false)?;
        t(100, &dave_email, &dave_uid, &dave_fpr, true)?;
        t(120, &dave_uid, &dave_uid, &dave_fpr, false)?;
        t(120, &dave_email, &dave_uid, &dave_fpr, false)?;

        // Not enough depth.
        t(100, &ellen_uid, &ellen_uid, &ellen_fpr, false)?;
        t(100, &ellen_email, &ellen_uid, &ellen_fpr, false)?;
        t(120, &ellen_uid, &ellen_uid, &ellen_fpr, false)?;
        t(120, &ellen_email, &ellen_uid, &ellen_fpr, false)?;

        // No such User ID on dave's key.
        t(100, &ellen_uid, &ellen_uid, &dave_fpr, false)?;
        t(100, &ellen_email, &ellen_uid, &dave_fpr, false)?;
        t(120, &ellen_uid, &ellen_uid, &dave_fpr, false)?;
        t(120, &ellen_email, &ellen_uid, &dave_fpr, false)?;

        // Normalized.
        t(100, &dave_email_uc1, &dave_uid, &dave_fpr, true)?;
        t(100, &dave_email_uc2, &dave_uid, &dave_fpr, true)?;


        // Puny code and case normalization.
        let alice_fpr: Fingerprint =
            "B8DA8B318149B1C8C0CBD1ECB1CEC6D3CD00E69D"
           .parse().expect("valid fingerprint");
        let alice_uid
            = UserID::from("<alice@example.org>");
        let alice_email
            = UserID::from("alice@example.org");

        let hans_fpr: Fingerprint =
            "74767C4F2B15F57F3394FCA99DE867E6CA6A2756"
           .parse().expect("valid fingerprint");
        let hans_uid
            = UserID::from("<hÄNS@bücher.tld>");
        // Certified by: B8DA8B318149B1C8C0CBD1ECB1CEC6D3CD00E69D

        let hans_email
            = UserID::from("hÄNS@bücher.tld");
        let hans_email_lowercase
            = UserID::from("häns@bücher.tld");
        let hans_email_punycode
            = UserID::from("hÄNS@xn--bcher-kva.tld");
        let hans_email_punycode_lowercase
            = UserID::from("häns@xn--bcher-kva.tld");

        let carol_fpr: Fingerprint =
            "7432C123761B94EC50D50CF6562B9ADEE7F789F6"
           .parse().expect("valid fingerprint");
        let carol_uid
            = UserID::from("<carol@example.org>");
        // Certified by: 74767C4F2B15F57F3394FCA99DE867E6CA6A2756

        let carol_email
            = UserID::from("carol@example.org");

        let t = |amount: usize, pattern: &UserID,
                 userid: &UserID, target: &Fingerprint, success: bool|
        {
            let output: &[(usize, &str)] = &[
                (1, &format!("[✓] {} {}: ", target, userid)),
                (1, "[✓] ")];
            test("puny-code.pgp", &alice_fpr,
                 &[], "authenticate", &["--email"],
                 amount, Some(pattern), Some(target),
                 success, output)
        };

        t(120, &alice_email, &alice_uid, &alice_fpr, true)?;

        t(100, &hans_email, &hans_uid, &hans_fpr, true)?;
        t(100, &hans_email_lowercase, &hans_uid, &hans_fpr, true)?;
        t(100, &hans_email_punycode, &hans_uid, &hans_fpr, true)?;
        t(100, &hans_email_punycode_lowercase, &hans_uid, &hans_fpr, true)?;

        t(100, &carol_email, &carol_uid, &carol_fpr, true)?;

        Ok(())
    }

    // Test looking up a certificate by User ID.
    #[test]
    #[allow(unused)]
    fn lookup() -> Result<()> {
        let alice_fpr: Fingerprint =
            "85DAB65713B2D0ABFC5A4F28BC10C9CE4A699D8D"
           .parse().expect("valid fingerprint");
        let alice_uid = UserID::from("<alice@example.org>");
        let alice_uid_uppercase = UserID::from("<ALICE@EXAMPLE.ORG>");
        let alice_uid_uppercase2 = UserID::from("<alice@EXAMPLE.ORG>");

        let dave_fpr: Fingerprint =
            "329D5AAF73DC70B4E3DD2D11677CB70FFBFE1281"
           .parse().expect("valid fingerprint");
        let dave_uid = UserID::from("<dave@example.org>");
        let dave_uid_uppercase = UserID::from("<DAVE@EXAMPLE.ORG>");
        let dave_uid_uppercase2 = UserID::from("<dave@EXAMPLE.ORG>");

        let ellen_fpr: Fingerprint =
            "A7319A9B166AB530A5FBAC8AB43CA77F7C176AF4"
           .parse().expect("valid fingerprint");
        let ellen_uid
            = UserID::from("<ellen@example.org>");
        // Certified by: 329D5AAF73DC70B4E3DD2D11677CB70FFBFE1281

        let t = |amount: usize, pattern: &UserID,
                 userid: &UserID, target: &Fingerprint,
                 success: bool|
        {
            let output: &[(usize, &str)] = &[
                (1, &format!("[✓] {} {}: ", target, userid)),
                (1, "[✓] ")];

            test("simple.pgp", &alice_fpr, &[], "lookup", &[],
                 amount, Some(pattern), None,
                 success, output)
        };

        t(100, &alice_uid, &alice_uid, &alice_fpr, true)?;
        t(120, &alice_uid, &alice_uid, &alice_fpr, true)?;

        t(100, &dave_uid, &dave_uid, &dave_fpr, true)?;
        t(120, &dave_uid, &dave_uid, &dave_fpr, false)?;

        // Not enough depth.
        t(100, &ellen_uid, &ellen_uid, &ellen_fpr, false)?;
        t(120, &ellen_uid, &ellen_uid, &ellen_fpr, false)?;

        // No such User ID.
        let gary = UserID::from("Gary <gary@some.org>");
        t(100, &gary, &gary, &dave_fpr, false)?;

        // We need an exact match.
        t(100, &alice_uid_uppercase, &alice_uid, &alice_fpr, false)?;
        t(100, &alice_uid_uppercase2, &alice_uid, &alice_fpr, false)?;
        t(100, &dave_uid_uppercase, &dave_uid, &dave_fpr, false)?;
        t(100, &dave_uid_uppercase2, &dave_uid, &dave_fpr, false)?;

        Ok(())
    }

    // Test looking up a certificate by email address.
    #[test]
    #[allow(unused)]
    fn lookup_email() -> Result<()> {
        let alice_fpr: Fingerprint =
            "85DAB65713B2D0ABFC5A4F28BC10C9CE4A699D8D"
           .parse().expect("valid fingerprint");
        let alice_uid = UserID::from("<alice@example.org>");
        let alice_email = UserID::from("alice@example.org");

        let dave_fpr: Fingerprint =
            "329D5AAF73DC70B4E3DD2D11677CB70FFBFE1281"
           .parse().expect("valid fingerprint");
        let dave_uid = UserID::from("<dave@example.org>");
        let dave_email = UserID::from("dave@example.org");

        let ellen_fpr: Fingerprint =
            "A7319A9B166AB530A5FBAC8AB43CA77F7C176AF4"
           .parse().expect("valid fingerprint");
        let ellen_uid
            = UserID::from("<ellen@example.org>");
        let ellen_email = UserID::from("ellen@example.org");
        // Certified by: 329D5AAF73DC70B4E3DD2D11677CB70FFBFE1281

        let t = |amount: usize, pattern: &UserID,
                 userid: &UserID, target: &Fingerprint,
                 success: bool|
        {
            let output: &[(usize, &str)] = &[
                (1, &format!("[✓] {} {}: ", target, userid)),
                (1, "[✓] ")];

            test("simple.pgp", &alice_fpr, &[], "lookup", &["--email"],
                 amount, Some(pattern), None,
                 success, output)
        };

        t(100, &alice_email, &alice_uid, &alice_fpr, true)?;
        t(120, &alice_email, &alice_uid, &alice_fpr, true)?;

        t(100, &dave_email, &dave_uid, &dave_fpr, true)?;
        t(120, &dave_email, &dave_uid, &dave_fpr, false)?;

        // Not enough depth.
        t(100, &ellen_email, &ellen_uid, &ellen_fpr, false)?;
        t(120, &ellen_email, &ellen_uid, &ellen_fpr, false)?;

        // No such User ID.
        let gary = UserID::from("gary@some.org");
        t(100, &gary, &gary, &dave_fpr, false)?;

        Ok(())
    }

    // Test identifying a certificate.
    #[test]
    #[allow(unused)]
    fn identify() -> Result<()> {
        let alice_fpr: Fingerprint =
            "2A2A4A23A7EEC119BC0B46642B3825DC02A05FEA"
           .parse().expect("valid fingerprint");
        let alice_uid
            = UserID::from("<alice@example.org>");

        let bob_fpr: Fingerprint =
            "03182611B91B1E7E20B848E83DFC151ABFAD85D5"
           .parse().expect("valid fingerprint");
        let bob_uid
            = UserID::from("<bob@other.org>");
        // Certified by: 2A2A4A23A7EEC119BC0B46642B3825DC02A05FEA
        let bob_some_org_uid
            = UserID::from("<bob@some.org>");
        // Certified by: 2A2A4A23A7EEC119BC0B46642B3825DC02A05FEA

        let carol_fpr: Fingerprint =
            "9CA36907B46FE7B6B9EE9601E78064C12B6D7902"
           .parse().expect("valid fingerprint");
        let carol_uid
            = UserID::from("<carol@example.org>");
        // Certified by: 03182611B91B1E7E20B848E83DFC151ABFAD85D5

        let dave_fpr: Fingerprint =
            "C1BC6794A6C6281B968A6A41ACE2055D610CEA03"
           .parse().expect("valid fingerprint");
        let dave_uid
            = UserID::from("<dave@other.org>");
        // Certified by: 9CA36907B46FE7B6B9EE9601E78064C12B6D7902


        let t = |amount: usize, userids: &[&UserID], target: &Fingerprint,
                 success: bool|
        {
            let output = userids
                .iter()
                .map(|userid| {
                    (1, format!("[✓] {} {}: ", target, userid))
                })
                .chain(vec![(userids.len(), String::from("[✓] "))].into_iter())
                .collect::<Vec<_>>();
            test("multiple-userids-1.pgp", &alice_fpr, &[], "identify", &[],
                 amount, None, Some(target),
                 success, &output)
        };

        t(100, &[&alice_uid], &alice_fpr, true)?;
        t(120, &[&alice_uid], &alice_fpr, true)?;

        t(50, &[&dave_uid], &dave_fpr, true)?;
        t(120, &[&dave_uid], &dave_fpr, false)?;

        t(50, &[&bob_uid, &bob_some_org_uid], &bob_fpr, true)?;
        t(120, &[&bob_uid, &bob_some_org_uid], &bob_fpr, false)?;

        Ok(())
    }

    // List all authenticated bindings.
    #[test]
    #[allow(unused)]
    fn list() -> Result<()> {
        let alice_fpr: Fingerprint =
            "85DAB65713B2D0ABFC5A4F28BC10C9CE4A699D8D"
           .parse().expect("valid fingerprint");
        let alice_uid
            = UserID::from("<alice@example.org>");

        let bob_fpr: Fingerprint =
            "39A479816C934B9E0464F1F4BC1DCFDEADA4EE90"
           .parse().expect("valid fingerprint");
        let bob_uid
            = UserID::from("<bob@example.org>");
        // Certified by: 85DAB65713B2D0ABFC5A4F28BC10C9CE4A699D8D

        let carol_fpr: Fingerprint =
            "43530F91B450EDB269AA58821A1CF4DC7F500F04"
           .parse().expect("valid fingerprint");
        let carol_uid
            = UserID::from("<carol@example.org>");
        // Certified by: 39A479816C934B9E0464F1F4BC1DCFDEADA4EE90

        let dave_fpr: Fingerprint =
            "329D5AAF73DC70B4E3DD2D11677CB70FFBFE1281"
           .parse().expect("valid fingerprint");
        let dave_uid
            = UserID::from("<dave@example.org>");
        // Certified by: 43530F91B450EDB269AA58821A1CF4DC7F500F04

        let ellen_fpr: Fingerprint =
            "A7319A9B166AB530A5FBAC8AB43CA77F7C176AF4"
           .parse().expect("valid fingerprint");
        let ellen_uid
            = UserID::from("<ellen@example.org>");
        // Certified by: 329D5AAF73DC70B4E3DD2D11677CB70FFBFE1281

        let frank_fpr: Fingerprint =
            "2693237D2CED0BB68F118D78DC86A97CD2C819D9"
           .parse().expect("valid fingerprint");
        let frank_uid
            = UserID::from("<frank@example.org>");

        let t = |amount: usize, bindings: &[(&UserID, &Fingerprint)],
                 success: bool|
        {
            let output = bindings
                .iter()
                .map(|(userid, target)| {
                    (1, format!("[✓] {} {}: ", target, userid))
                })
                .chain(vec![(bindings.len(), String::from("[✓] "))].into_iter())
                .collect::<Vec<_>>();
            test("simple.pgp", &alice_fpr, &[], "list", &[],
                 amount, None, None,
                 success, &output)
        };

        t(120,
          &[(&alice_uid, &alice_fpr)],
          true)?;

        t(50,
          &[(&alice_uid, &alice_fpr),
            (&bob_uid, &bob_fpr),
            (&carol_uid, &carol_fpr),
            (&dave_uid, &dave_fpr),
          ],
          true)?;

        Ok(())
    }

    // List all authenticated bindings matching a pattern.
    #[test]
    #[allow(unused)]
    fn list_pattern() -> Result<()> {
        let alice_fpr: Fingerprint =
            "2A2A4A23A7EEC119BC0B46642B3825DC02A05FEA"
           .parse().expect("valid fingerprint");
        let alice_uid
            = UserID::from("<alice@example.org>");

        let bob_fpr: Fingerprint =
            "03182611B91B1E7E20B848E83DFC151ABFAD85D5"
           .parse().expect("valid fingerprint");
        let bob_uid
            = UserID::from("<bob@other.org>");
        // Certified by: 2A2A4A23A7EEC119BC0B46642B3825DC02A05FEA
        let bob_some_org_uid
            = UserID::from("<bob@some.org>");
        // Certified by: 2A2A4A23A7EEC119BC0B46642B3825DC02A05FEA

        let carol_fpr: Fingerprint =
            "9CA36907B46FE7B6B9EE9601E78064C12B6D7902"
           .parse().expect("valid fingerprint");
        let carol_uid
            = UserID::from("<carol@example.org>");
        // Certified by: 03182611B91B1E7E20B848E83DFC151ABFAD85D5

        let dave_fpr: Fingerprint =
            "C1BC6794A6C6281B968A6A41ACE2055D610CEA03"
           .parse().expect("valid fingerprint");
        let dave_uid
            = UserID::from("<dave@other.org>");
        // Certified by: 9CA36907B46FE7B6B9EE9601E78064C12B6D7902


        let t = |amount: usize, pattern: &str,
                 bindings: &[(&UserID, &Fingerprint)],
                 success: bool|
        {
            let output = bindings
                .iter()
                .map(|(userid, target)| {
                    (1, format!("[✓] {} {}: ", target, userid))
                })
                .chain(vec![(bindings.len(), String::from("[✓] "))].into_iter())
                .collect::<Vec<_>>();
            test("multiple-userids-1.pgp", &alice_fpr, &[], "list", &[pattern],
                 amount, None, None,
                 success, &output)
        };

        t(50, "bob",
          &[(&bob_uid, &bob_fpr),
            (&bob_some_org_uid, &bob_fpr)
          ],
          true)?;

        t(50, "BOB",
          &[(&bob_uid, &bob_fpr),
            (&bob_some_org_uid, &bob_fpr)
          ],
          true)?;

        t(50, "@example.org",
          &[(&alice_uid, &alice_fpr),
            (&carol_uid, &carol_fpr),
          ],
          true)?;

        t(50, "@EXAMPLE.ORG",
          &[(&alice_uid, &alice_fpr),
            (&carol_uid, &carol_fpr),
          ],
          true)?;

        t(50, "@OTHER.ORG",
          &[(&bob_uid, &bob_fpr),
            (&dave_uid, &dave_fpr),
          ],
          true)?;

        t(50, "ORG",
          &[(&alice_uid, &alice_fpr),
            (&bob_uid, &bob_fpr),
            (&bob_some_org_uid, &bob_fpr),
            (&carol_uid, &carol_fpr),
            (&dave_uid, &dave_fpr),
          ],
          true)?;


        // Puny code.
        let alice_fpr: Fingerprint =
            "B8DA8B318149B1C8C0CBD1ECB1CEC6D3CD00E69D"
           .parse().expect("valid fingerprint");
        let alice_uid
            = UserID::from("<alice@example.org>");
        let alice_email
            = UserID::from("alice@example.org");

        let hans_fpr: Fingerprint =
            "74767C4F2B15F57F3394FCA99DE867E6CA6A2756"
           .parse().expect("valid fingerprint");
        let hans_uid
            = UserID::from("<hÄNS@bücher.tld>");
        // Certified by: B8DA8B318149B1C8C0CBD1ECB1CEC6D3CD00E69D

        let hans_email = "hÄNS@bücher.tld";
        let hans_email_punycode = "hÄNS@xn--bcher-kva.tld";
        let hans_email_punycode_lowercase = "häns@xn--bcher-kva.tld";

        let carol_fpr: Fingerprint =
            "7432C123761B94EC50D50CF6562B9ADEE7F789F6"
           .parse().expect("valid fingerprint");
        let carol_uid
            = UserID::from("<carol@example.org>");
        // Certified by: 74767C4F2B15F57F3394FCA99DE867E6CA6A2756

        let carol_email
            = UserID::from("carol@example.org");

        let t = |amount: usize, pattern: &str,
                 bindings: &[(&UserID, &Fingerprint)],
                 success: bool|
        {
            let output = bindings
                .iter()
                .map(|(userid, target)| {
                    (1, format!("[✓] {} {}: ", target, userid))
                })
                .chain(vec![(bindings.len(), String::from("[✓] "))].into_iter())
                .collect::<Vec<_>>();
            test("puny-code.pgp", &alice_fpr, &[], "list", &[pattern],
                 amount, None, None,
                 success, &output)
        };

        // If we don't provide --email, then we only case
        // insensitively match on the raw User ID; we don't perform
        // puny code normalization.
        t(100, "bücher.tld",
          &[(&hans_uid, &hans_fpr)],
          true)?;

        t(100, "BÜCHER.TLD",
          &[(&hans_uid, &hans_fpr)],
          true)?;

        t(100, hans_email,
          &[(&hans_uid, &hans_fpr)],
          true)?;

        t(100, &format!("<{}>", hans_email),
          &[(&hans_uid, &hans_fpr)],
          true)?;

        t(100, hans_email_punycode,
          &[(&hans_uid, &hans_fpr)],
          false)?;

        t(100, hans_email_punycode_lowercase,
          &[(&hans_uid, &hans_fpr)],
          false)?;


        Ok(())
    }

    // List all authenticated bindings where the email address matches
    // a pattern.
    #[test]
    #[allow(unused)]
    fn list_email_pattern() -> Result<()> {
        // Puny code and case normalization.
        let alice_fpr: Fingerprint =
            "B8DA8B318149B1C8C0CBD1ECB1CEC6D3CD00E69D"
           .parse().expect("valid fingerprint");
        let alice_uid
            = UserID::from("<alice@example.org>");
        let alice_email
            = UserID::from("alice@example.org");

        let hans_fpr: Fingerprint =
            "74767C4F2B15F57F3394FCA99DE867E6CA6A2756"
           .parse().expect("valid fingerprint");
        let hans_uid
            = UserID::from("<hÄNS@bücher.tld>");
        // Certified by: B8DA8B318149B1C8C0CBD1ECB1CEC6D3CD00E69D

        let hans_email = "hÄNS@bücher.tld";
        let hans_email_punycode = "hÄNS@xn--bcher-kva.tld";
        let hans_email_punycode_lowercase = "häns@xn--bcher-kva.tld";

        let carol_fpr: Fingerprint =
            "7432C123761B94EC50D50CF6562B9ADEE7F789F6"
           .parse().expect("valid fingerprint");
        let carol_uid
            = UserID::from("<carol@example.org>");
        // Certified by: 74767C4F2B15F57F3394FCA99DE867E6CA6A2756

        let carol_email
            = UserID::from("carol@example.org");

        let t = |amount: usize, pattern: &str,
                 bindings: &[(&UserID, &Fingerprint)],
                 success: bool|
        {
            let output = bindings
                .iter()
                .map(|(userid, target)| {
                    (1, format!("[✓] {} {}: ", target, userid))
                })
                .chain(vec![(bindings.len(), String::from("[✓] "))].into_iter())
                .collect::<Vec<_>>();
            test("puny-code.pgp", &alice_fpr,
                 &[], "list", &["--email", pattern],
                 amount, None, None,
                 success, &output)
        };

        t(100, "bücher.tld",
          &[(&hans_uid, &hans_fpr)],
          true)?;

        t(100, "BÜCHER.TLD",
          &[(&hans_uid, &hans_fpr)],
          true)?;

        t(100, hans_email,
          &[(&hans_uid, &hans_fpr)],
          true)?;

        t(100, &format!("<{}>", hans_email),
          &[(&hans_uid, &hans_fpr)],
          false)?;

        t(100, hans_email_punycode,
          &[(&hans_uid, &hans_fpr)],
          true)?;

        t(100, hans_email_punycode_lowercase,
          &[(&hans_uid, &hans_fpr)],
          true)?;

        Ok(())
    }

    #[test]
    #[allow(unused)]
    fn path_simple() -> Result<()> {
        let alice_fpr: Fingerprint =
            "85DAB65713B2D0ABFC5A4F28BC10C9CE4A699D8D"
           .parse().expect("valid fingerprint");
        let alice_uid
            = UserID::from("<alice@example.org>");

        let bob_fpr: Fingerprint =
            "39A479816C934B9E0464F1F4BC1DCFDEADA4EE90"
           .parse().expect("valid fingerprint");
        let bob_uid
            = UserID::from("<bob@example.org>");
        // Certified by: 85DAB65713B2D0ABFC5A4F28BC10C9CE4A699D8D

        let carol_fpr: Fingerprint =
            "43530F91B450EDB269AA58821A1CF4DC7F500F04"
           .parse().expect("valid fingerprint");
        let carol_uid
            = UserID::from("<carol@example.org>");
        // Certified by: 39A479816C934B9E0464F1F4BC1DCFDEADA4EE90

        let dave_fpr: Fingerprint =
            "329D5AAF73DC70B4E3DD2D11677CB70FFBFE1281"
           .parse().expect("valid fingerprint");
        let dave_uid
            = UserID::from("<dave@example.org>");
        // Certified by: 43530F91B450EDB269AA58821A1CF4DC7F500F04

        let ellen_fpr: Fingerprint =
            "A7319A9B166AB530A5FBAC8AB43CA77F7C176AF4"
           .parse().expect("valid fingerprint");
        let ellen_uid
            = UserID::from("<ellen@example.org>");
        // Certified by: 329D5AAF73DC70B4E3DD2D11677CB70FFBFE1281

        let frank_fpr: Fingerprint =
            "2693237D2CED0BB68F118D78DC86A97CD2C819D9"
           .parse().expect("valid fingerprint");
        let frank_uid
            = UserID::from("<frank@example.org>");

        // Alice certifies Bob at trust amount = 100. (120 required).
        test("simple.pgp",
             None, // No trust root for path.
             &[], "path",
             &[ &alice_fpr.to_string(),
                 &bob_fpr.to_string(),
                 &String::from_utf8_lossy(bob_uid.value()),
             ],
             120, None, None,
             false,
             &[
                 (1, "[ ] ")
             ])
            .expect("valid");

        // As above, but we only require 100.
        test("simple.pgp",
             None, // No trust root for path.
             &[], "path",
             &[ &alice_fpr.to_string(),
                 &bob_fpr.to_string(),
                 &String::from_utf8_lossy(bob_uid.value()),
             ],
             100, None, None,
             true,
             &[
                 (1, "[✓] ")
             ])
            .expect("valid");

        // Alice makes Bob a level 2 trusted introducer.
        // Bob certificates Carol, but for Bob.
        test("simple.pgp",
             None, // No trust root for path.
             &[], "path",
             &[ &alice_fpr.to_string(),
                 &bob_fpr.to_string(),
                 &carol_fpr.to_string(),
                 &String::from_utf8_lossy(bob_uid.value()),
             ],
             100, None, None,
             false,
             &[
                 (1, "[ ] ")
             ])
            .expect("valid");

        // Alice makes Bob a level 2 trusted introducer.
        // Bob certificates Carol.
        test("simple.pgp",
             None, // No trust root for path.
             &[], "path",
             &[ &alice_fpr.to_string(),
                 &bob_fpr.to_string(),
                 &carol_fpr.to_string(),
                 &String::from_utf8_lossy(carol_uid.value()),
             ],
             100, None, None,
             true,
             &[
                 (1, "[✓] ")
             ])
            .expect("valid");

        // Alice makes Bob a level 2 trusted introducer.
        // Bob makes Carol a level 1 trust introducer.
        // Carol certifies Dave.
        test("simple.pgp",
             None, // No trust root for path.
             &[], "path",
             &[ &alice_fpr.to_string(),
                 &bob_fpr.to_string(),
                 &carol_fpr.to_string(),
                 &dave_fpr.to_string(),
                 &String::from_utf8_lossy(dave_uid.value()),
             ],
             100, None, None,
             true,
             &[
                 (1, "[✓] ")
             ])
            .expect("valid");

        // Alice makes Bob a level 2 trusted introducer (require level 3).
        // Bob makes Carol a level 1 trusted introducer (require level 2).
        // Carol makes Dave a level 1 trusted introducer.
        // Dave certifies Ellen.
        test("simple.pgp",
             None, // No trust root for path.
             &[], "path",
             &[ &alice_fpr.to_string(),
                 &bob_fpr.to_string(),
                 &carol_fpr.to_string(),
                 &dave_fpr.to_string(),
                 &ellen_fpr.to_string(),
                 &String::from_utf8_lossy(ellen_uid.value()),
             ],
             100, None, None,
             false,
             &[
                 (1, "[ ] ")
             ])
            .expect("valid");

        // Alice makes Bob a level 2 trusted introducer.
        // Bob does *not* certify Dave.
        // Dave certifies Ellen.
        test("simple.pgp",
             None, // No trust root for path.
             &[], "path",
             &[ &alice_fpr.to_string(),
                 &bob_fpr.to_string(),
                 &dave_fpr.to_string(),
                 &ellen_fpr.to_string(),
                 &String::from_utf8_lossy(ellen_uid.value()),
             ],
             100, None, None,
             false,
             &[
                 (1, "[ ] ")
             ])
            .expect("valid");

        Ok(())
    }

    #[test]
    #[allow(unused)]
    fn path_missing_certs() -> Result<()> {
        let alice_fpr: Fingerprint =
            "85DAB65713B2D0ABFC5A4F28BC10C9CE4A699D8D"
           .parse().expect("valid fingerprint");
        let alice_uid
            = UserID::from("<alice@example.org>");

        let bob_fpr: Fingerprint =
            "39A479816C934B9E0464F1F4BC1DCFDEADA4EE90"
           .parse().expect("valid fingerprint");
        let bob_uid
            = UserID::from("<bob@example.org>");
        // Certified by: 85DAB65713B2D0ABFC5A4F28BC10C9CE4A699D8D

        let carol_fpr: Fingerprint =
            "43530F91B450EDB269AA58821A1CF4DC7F500F04"
           .parse().expect("valid fingerprint");
        let carol_uid
            = UserID::from("<carol@example.org>");
        // Certified by: 39A479816C934B9E0464F1F4BC1DCFDEADA4EE90

        let dave_fpr: Fingerprint =
            "329D5AAF73DC70B4E3DD2D11677CB70FFBFE1281"
           .parse().expect("valid fingerprint");
        let dave_uid
            = UserID::from("<dave@example.org>");
        // Certified by: 43530F91B450EDB269AA58821A1CF4DC7F500F04

        let ellen_fpr: Fingerprint =
            "A7319A9B166AB530A5FBAC8AB43CA77F7C176AF4"
           .parse().expect("valid fingerprint");
        let ellen_uid
            = UserID::from("<ellen@example.org>");
        // Certified by: 329D5AAF73DC70B4E3DD2D11677CB70FFBFE1281

        let frank_fpr: Fingerprint =
            "2693237D2CED0BB68F118D78DC86A97CD2C819D9"
           .parse().expect("valid fingerprint");
        let frank_uid
            = UserID::from("<frank@example.org>");

        let missing_fpr: Fingerprint =
            "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA"
           .parse().expect("valid fingerprint");

        // Alice tsigns Bob at depth = 2, trust amount = 100.
        // Bob certifies Carol, trust amount = 100.
        test("simple.pgp",
             None, // No trust root for path.
             &[], "path",
             &[ &alice_fpr.to_string(),
                 &bob_fpr.to_string(),
                 &carol_fpr.to_string(),
                 &String::from_utf8_lossy(carol_uid.value()),
             ],
             100, None, None,
             true,
             &[
                 (1, "[✓] ")
             ])
            .expect("valid");

        test("simple.pgp",
             None, // No trust root for path.
             &[], "path",
             &[ &missing_fpr.to_string(),
                 &bob_fpr.to_string(),
                 &carol_fpr.to_string(),
                 &String::from_utf8_lossy(carol_uid.value()),
             ],
             100, None, None,
             false,
             &[
                 (1, "[ ] ")
             ])
            .expect("valid");

        test("simple.pgp",
             None, // No trust root for path.
             &[], "path",
             &[ &alice_fpr.to_string(),
                 &missing_fpr.to_string(),
                 &carol_fpr.to_string(),
                 &String::from_utf8_lossy(carol_uid.value()),
             ],
             100, None, None,
             false,
             &[
                 (1, "[ ] ")
             ])
            .expect("valid");

        test("simple.pgp",
             None, // No trust root for path.
             &[], "path",
             &[ &alice_fpr.to_string(),
                 &bob_fpr.to_string(),
                 &missing_fpr.to_string(),
                 &String::from_utf8_lossy(carol_uid.value()),
             ],
             100, None, None,
             false,
             &[
                 (1, "[ ] ")
             ])
            .expect("valid");

        Ok(())
    }

    #[test]
    #[allow(unused)]
    fn path_singleton() -> Result<()> {
        let alice_fpr: Fingerprint =
            "85DAB65713B2D0ABFC5A4F28BC10C9CE4A699D8D"
           .parse().expect("valid fingerprint");
        let alice_uid
            = UserID::from("<alice@example.org>");

        let bob_fpr: Fingerprint =
            "39A479816C934B9E0464F1F4BC1DCFDEADA4EE90"
           .parse().expect("valid fingerprint");
        let bob_uid
            = UserID::from("<bob@example.org>");
        // Certified by: 85DAB65713B2D0ABFC5A4F28BC10C9CE4A699D8D

        let carol_fpr: Fingerprint =
            "43530F91B450EDB269AA58821A1CF4DC7F500F04"
           .parse().expect("valid fingerprint");
        let carol_uid
            = UserID::from("<carol@example.org>");
        // Certified by: 39A479816C934B9E0464F1F4BC1DCFDEADA4EE90

        let dave_fpr: Fingerprint =
            "329D5AAF73DC70B4E3DD2D11677CB70FFBFE1281"
           .parse().expect("valid fingerprint");
        let dave_uid
            = UserID::from("<dave@example.org>");
        // Certified by: 43530F91B450EDB269AA58821A1CF4DC7F500F04

        let ellen_fpr: Fingerprint =
            "A7319A9B166AB530A5FBAC8AB43CA77F7C176AF4"
           .parse().expect("valid fingerprint");
        let ellen_uid
            = UserID::from("<ellen@example.org>");
        // Certified by: 329D5AAF73DC70B4E3DD2D11677CB70FFBFE1281

        let frank_fpr: Fingerprint =
            "2693237D2CED0BB68F118D78DC86A97CD2C819D9"
           .parse().expect("valid fingerprint");
        let frank_uid
            = UserID::from("<frank@example.org>");

        let missing_fpr: Fingerprint =
            "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA"
           .parse().expect("valid fingerprint");

        // A self signed User ID.
        test("simple.pgp",
             None, // No trust root for path.
             &[], "path",
             &[ &alice_fpr.to_string(),
                 &String::from_utf8_lossy(alice_uid.value()),
             ],
             120, None, None,
             true,
             &[
                 (1, "[✓] ")
             ])
            .expect("valid");

        // A User ID that is not self signed.
        test("simple.pgp",
             None, // No trust root for path.
             &[], "path",
             &[ &alice_fpr.to_string(),
                 &String::from_utf8_lossy(bob_uid.value()),
             ],
             120, None, None,
             false,
             &[
                 (1, "[ ] ")
             ])
            .expect("valid");

        Ok(())
    }

    #[test]
    #[allow(unused)]
    fn path_multiple_userids_1() -> Result<()> {
        let alice_fpr: Fingerprint =
            "2A2A4A23A7EEC119BC0B46642B3825DC02A05FEA"
           .parse().expect("valid fingerprint");
        let alice_uid
            = UserID::from("<alice@example.org>");

        let bob_fpr: Fingerprint =
            "03182611B91B1E7E20B848E83DFC151ABFAD85D5"
           .parse().expect("valid fingerprint");
        let bob_uid
            = UserID::from("<bob@other.org>");
        // Certified by: 2A2A4A23A7EEC119BC0B46642B3825DC02A05FEA
        let bob_some_org_uid
            = UserID::from("<bob@some.org>");
        // Certified by: 2A2A4A23A7EEC119BC0B46642B3825DC02A05FEA

        let carol_fpr: Fingerprint =
            "9CA36907B46FE7B6B9EE9601E78064C12B6D7902"
           .parse().expect("valid fingerprint");
        let carol_uid
            = UserID::from("<carol@example.org>");
        // Certified by: 03182611B91B1E7E20B848E83DFC151ABFAD85D5

        let dave_fpr: Fingerprint =
            "C1BC6794A6C6281B968A6A41ACE2055D610CEA03"
           .parse().expect("valid fingerprint");
        let dave_uid
            = UserID::from("<dave@other.org>");
        // Certified by: 9CA36907B46FE7B6B9EE9601E78064C12B6D7902

        // Alice certifies Bob as:
        //   a level 2 trusted introducer, amount = 50
        //   a level 1 trusted introducer, amount = 70
        // Bob certifies Carol as a level 2 trusted introducer
        // Carol certifies Dave
        test("multiple-userids-1.pgp",
             None, // No trust root for path.
             &[], "path",
             &[ &alice_fpr.to_string(),
                 &bob_fpr.to_string(),
                 &carol_fpr.to_string(),
                 &dave_fpr.to_string(),
                 &String::from_utf8_lossy(dave_uid.value()),
             ],
             70, None, None,
             false,
             &[
                 (1, "[ ] ")
             ])
            .expect("valid");

        // As above, but
        test("multiple-userids-1.pgp",
             None, // No trust root for path.
             &[], "path",
             &[ &alice_fpr.to_string(),
                 &bob_fpr.to_string(),
                 &carol_fpr.to_string(),
                 &dave_fpr.to_string(),
                 &String::from_utf8_lossy(dave_uid.value()),
             ],
             50, None, None,
             true,
             &[
                 (1, "[✓] ")
             ])
            .expect("valid");
        Ok(())
    }

    #[test]
    #[allow(unused)]
    fn path_multiple_users_2() -> Result<()> {
        // Note: this also tests regular expressions.

        let alice_fpr: Fingerprint =
            "F1C99C4019837703DD17C45440F8A0141DF278EA"
           .parse().expect("valid fingerprint");
        let alice_uid
            = UserID::from("<alice@example.org>");

        let bob_fpr: Fingerprint =
            "5528B9E5DAFC519ED2E37F0377B332E4111456CB"
           .parse().expect("valid fingerprint");
        let bob_uid
            = UserID::from("<bob@other.org>");
        // Certified by: F1C99C4019837703DD17C45440F8A0141DF278EA
        let bob_some_org_uid
            = UserID::from("<bob@some.org>");
        // Certified by: F1C99C4019837703DD17C45440F8A0141DF278EA

        let carol_fpr: Fingerprint =
            "6F8291428420AB53576BAB4BEFF6477D3E348D71"
           .parse().expect("valid fingerprint");
        let carol_uid
            = UserID::from("<carol@example.org>");
        // Certified by: 5528B9E5DAFC519ED2E37F0377B332E4111456CB

        let dave_fpr: Fingerprint =
            "62C57D90DAD253DEA01D5A86C7382FD6285C18F0"
           .parse().expect("valid fingerprint");
        let dave_uid
            = UserID::from("<dave@other.org>");
        // Certified by: 6F8291428420AB53576BAB4BEFF6477D3E348D71

        let ed_fpr: Fingerprint =
            "0E974D0ACBA0C4D8F51D7CF68F048FF83B173504"
           .parse().expect("valid fingerprint");
        let ed_uid
            = UserID::from("<ed@example.org>");
        // Certified by: 6F8291428420AB53576BAB4BEFF6477D3E348D71

        let frank_fpr: Fingerprint =
            "5BEE3D41F85B2FCBC300DE4E18CB2BDA65465F03"
           .parse().expect("valid fingerprint");
        let frank_uid
            = UserID::from("<frank@other.org>");
        // Certified by: 5528B9E5DAFC519ED2E37F0377B332E4111456CB


        // Alice certifies Bob as:
        //   a level 255 trusted introducer, amount = 70 for other.org
        //   a level 1 trusted introducer, amount = 50
        // Bob certifies Frank@other.org
        test("multiple-userids-2.pgp",
             None, // No trust root for path.
             &[], "path",
             &[ &alice_fpr.to_string(),
                 &bob_fpr.to_string(),
                 &frank_fpr.to_string(),
                 &String::from_utf8_lossy(frank_uid.value()),
             ],
             70, None, None,
             true,
             &[
                 (1, "[✓] ")
             ])
            .expect("valid");

        // Alice certifies Bob as:
        //   a level 255 trusted introducer, amount = 70 for other.org
        //   a level 1 trusted introducer, amount = 50
        // Bob certifies Carol@example.org
        test("multiple-userids-2.pgp",
             None, // No trust root for path.
             &[], "path",
             &[ &alice_fpr.to_string(),
                 &bob_fpr.to_string(),
                 &carol_fpr.to_string(),
                 &String::from_utf8_lossy(carol_uid.value()),
             ],
             70, None, None,
             false,
             &[
                 (1, "[ ] ")
             ])
            .expect("valid");

        // As above, but reduce the required trust amount to 50.
        test("multiple-userids-2.pgp",
             None, // No trust root for path.
             &[], "path",
             &[ &alice_fpr.to_string(),
                 &bob_fpr.to_string(),
                 &carol_fpr.to_string(),
                 &String::from_utf8_lossy(carol_uid.value()),
             ],
             50, None, None,
             true,
             &[
                 (1, "[✓] ")
             ])
            .expect("valid");

        // Alice certifies Bob as:
        //   a level 255 trusted introducer, amount = 70 for other.org
        //   a level 1 trusted introducer, amount = 50
        // Bob certifies carol as a level 2 trusted introducer
        // Carol certifies dave@other.org
        test("multiple-userids-2.pgp",
             None, // No trust root for path.
             &[], "path",
             &[ &alice_fpr.to_string(),
                 &bob_fpr.to_string(),
                 &carol_fpr.to_string(),
                 &dave_fpr.to_string(),
                 &String::from_utf8_lossy(dave_uid.value()),
             ],
             70, None, None,
             true,
             &[
                 (1, "[✓] ")
             ])
            .expect("valid");

        // Alice certifies Bob as:
        //   a level 255 trusted introducer, amount = 70 for other.org
        //   a level 1 trusted introducer, amount = 50
        // Bob certifies carol as a level 2 trusted introducer
        // Carol certifies ed@example.org
        test("multiple-userids-2.pgp",
             None, // No trust root for path.
             &[], "path",
             &[ &alice_fpr.to_string(),
                 &bob_fpr.to_string(),
                 &carol_fpr.to_string(),
                 &ed_fpr.to_string(),
                 &String::from_utf8_lossy(ed_uid.value()),
             ],
             70, None, None,
             false,
             &[
                 (1, "[ ] ")
             ])
            .expect("valid");

        Ok(())
    }

    #[test]
    #[allow(unused)]
    fn path_sha1() -> Result<()> {
        let alice_fpr: Fingerprint =
            "B5FA089BA76FE3E17DC11660960E53286738F94C"
           .parse().expect("valid fingerprint");
        let alice_uid
            = UserID::from("<alice@example.org>");

        let bob_fpr: Fingerprint =
            "231BC4AB9D8CAB86D1622CE02C0CE554998EECDB"
           .parse().expect("valid fingerprint");
        let bob_uid
            = UserID::from("<bob@example.org>");
        // Certified by: B5FA089BA76FE3E17DC11660960E53286738F94C
        // Certified by: B5FA089BA76FE3E17DC11660960E53286738F94C

        let carol_fpr: Fingerprint =
            "FABA8485B2D4D5BF1582AA963A8115E774FA9852"
           .parse().expect("valid fingerprint");
        let carol_uid
            = UserID::from("<carol@example.org>");
        // Certified by: 231BC4AB9D8CAB86D1622CE02C0CE554998EECDB
        // Certified by: 231BC4AB9D8CAB86D1622CE02C0CE554998EECDB

        // Alice certifies Bob as
        //   a level 1 trusted introducer, amount = 120 using sha1
        //   a level 1 trusted introducer, amount = 60 using sha512 (future)
        // Bob certifies carol as a
        //   a level 1 trusted introducer, amount = 120 using sha1
        //   a level 1 trusted introducer, amount = 60 using sha512 (future)
        //
        // The valid signatures won't be used because they are from
        // the future (after the reference time).
        test("sha1.pgp",
             None, // No trust root for path.
             &[ "--time", "2023-01-10T15:07:01" ],
             "path",
             &[ &alice_fpr.to_string(),
                 &bob_fpr.to_string(),
                 &carol_fpr.to_string(),
                 &String::from_utf8_lossy(carol_uid.value()),
             ],
             120, None, None,
             false,
             &[
                 (1, "[ ] ")
             ])
            .expect("valid");

        // Again, but this time only require a trust amount of 60.
        test("sha1.pgp",
             None, // No trust root for path.
             &[ "--time", "2023-01-10T15:07:01" ],
             "path",
             &[ &alice_fpr.to_string(),
                 &bob_fpr.to_string(),
                 &carol_fpr.to_string(),
                 &String::from_utf8_lossy(carol_uid.value()),
             ],
             60, None, None,
             false,
             &[
                 (1, "[ ] ")
             ])
            .expect("valid");

        // Again, after the SHA256 certificates are valid.  But with a
        // trust amount of 120.
        test("sha1.pgp",
             None, // No trust root for path.
             &[],
             "path",
             &[ &alice_fpr.to_string(),
                 &bob_fpr.to_string(),
                 &carol_fpr.to_string(),
                 &String::from_utf8_lossy(carol_uid.value()),
             ],
             120, None, None,
             false,
             &[
                 (1, "[ ] ")
             ])
            .expect("valid");

        test("sha1.pgp",
             None, // No trust root for path.
             &[],
             "path",
             &[ &alice_fpr.to_string(),
                 &bob_fpr.to_string(),
                 &carol_fpr.to_string(),
                 &String::from_utf8_lossy(carol_uid.value()),
             ],
             60, None, None,
             true,
             &[
                 (1, "[✓] ")
             ])
            .expect("valid");

        Ok(())
    }

    #[test]
    #[allow(unused)]
    fn authenticate_certification_network_simple() -> Result<()> {
        let alice_fpr: Fingerprint =
            "85DAB65713B2D0ABFC5A4F28BC10C9CE4A699D8D"
           .parse().expect("valid fingerprint");
        let alice_uid
            = UserID::from("<alice@example.org>");

        let bob_fpr: Fingerprint =
            "39A479816C934B9E0464F1F4BC1DCFDEADA4EE90"
           .parse().expect("valid fingerprint");
        let bob_uid
            = UserID::from("<bob@example.org>");
        // Certified by: 85DAB65713B2D0ABFC5A4F28BC10C9CE4A699D8D

        let carol_fpr: Fingerprint =
            "43530F91B450EDB269AA58821A1CF4DC7F500F04"
           .parse().expect("valid fingerprint");
        let carol_uid
            = UserID::from("<carol@example.org>");
        // Certified by: 39A479816C934B9E0464F1F4BC1DCFDEADA4EE90

        let dave_fpr: Fingerprint =
            "329D5AAF73DC70B4E3DD2D11677CB70FFBFE1281"
           .parse().expect("valid fingerprint");
        let dave_uid
            = UserID::from("<dave@example.org>");
        // Certified by: 43530F91B450EDB269AA58821A1CF4DC7F500F04

        let ellen_fpr: Fingerprint =
            "A7319A9B166AB530A5FBAC8AB43CA77F7C176AF4"
           .parse().expect("valid fingerprint");
        let ellen_uid
            = UserID::from("<ellen@example.org>");
        // Certified by: 329D5AAF73DC70B4E3DD2D11677CB70FFBFE1281

        let frank_fpr: Fingerprint =
            "2693237D2CED0BB68F118D78DC86A97CD2C819D9"
           .parse().expect("valid fingerprint");
        let frank_uid
            = UserID::from("<frank@example.org>");

        test("simple.pgp",
             &alice_fpr,
             &[], "authenticate", &[],
             100, Some(&ellen_uid), Some(&ellen_fpr),
             false,
             &[
                 (1, "[ ] ")
             ])
            .expect("valid");

        test("simple.pgp",
             &alice_fpr,
             &["--certification-network"], "authenticate", &[],
             100, Some(&ellen_uid), Some(&ellen_fpr),
             true,
             &[
                 (1, "[✓] ")
             ])
            .expect("valid");

        Ok(())
    }

    #[test]
    #[allow(unused)]
    fn authenticate_certification_network() -> Result<()> {
        let alice_fpr: Fingerprint =
            "B2B371214EF71AFD16E42C62D81360B4C0489225"
           .parse().expect("valid fingerprint");
        let alice_uid
            = UserID::from("<alice@example.org>");
        // Certified by: 9A1AE937B5CB8BC46048AB63023CC01973ED9DF3

        let bob_fpr: Fingerprint =
            "A68DF00EB82F9C49C27CC7723C5F5BBE6B790C05"
           .parse().expect("valid fingerprint");
        let bob_uid
            = UserID::from("<bob@example.org>");
        // Certified by: B2B371214EF71AFD16E42C62D81360B4C0489225

        let carol_fpr: Fingerprint =
            "AB9EF1C89631519842ED559697557DD147D99C97"
           .parse().expect("valid fingerprint");
        let carol_uid
            = UserID::from("<carol@example.org>");
        // Certified by: A68DF00EB82F9C49C27CC7723C5F5BBE6B790C05

        let dave_fpr: Fingerprint =
            "9A1AE937B5CB8BC46048AB63023CC01973ED9DF3"
           .parse().expect("valid fingerprint");
        let dave_uid
            = UserID::from("<dave@example.org>");
        // Certified by: AB9EF1C89631519842ED559697557DD147D99C97

        // Alice does not make Bob a trusted introducer.  So without
        // --certificate-network, she can only authenticate Bob, but
        // not Carol or Dave.
        test("certification-network.pgp",
             &alice_fpr,
             &[], "authenticate", &[],
             120, Some(&bob_uid), Some(&bob_fpr),
             true,
             &[
                 (1, "[✓] ")
             ])
            .expect("valid");

        test("certification-network.pgp",
             &alice_fpr,
             &[], "authenticate", &[],
             120, Some(&carol_uid), Some(&carol_fpr),
             false,
             &[
                 (1, "[ ] ")
             ])
            .expect("valid");

        test("certification-network.pgp",
             &alice_fpr,
             &[], "authenticate", &[],
             120, Some(&dave_uid), Some(&dave_fpr),
             false,
             &[
                 (1, "[ ] ")
             ])
            .expect("valid");

        // With --certification-network, she can authenticate them all.
        test("certification-network.pgp",
             &alice_fpr,
             &["--certification-network"], "authenticate", &[],
             120, Some(&bob_uid), Some(&bob_fpr),
             true,
             &[
                 (1, "[✓] ")
             ])
            .expect("valid");

        test("certification-network.pgp",
             &alice_fpr,
             &["--certification-network"], "authenticate", &[],
             120, Some(&carol_uid), Some(&carol_fpr),
             true,
             &[
                 (1, "[✓] ")
             ])
            .expect("valid");

        test("certification-network.pgp",
             &alice_fpr,
             &["--certification-network"], "authenticate", &[],
             120, Some(&dave_uid), Some(&dave_fpr),
             true,
             &[
                 (1, "[✓] ")
             ])
            .expect("valid");

        // dave authenticates alice for 60 of 120.
        test("certification-network.pgp",
             &dave_fpr,
             &[], "authenticate", &[],
             120, Some(&alice_uid), Some(&alice_fpr),
             false,
             &[
                 (1, "[ ] ")
             ])
            .expect("valid");

        test("certification-network.pgp",
             &dave_fpr,
             &[], "authenticate", &[],
             60, Some(&alice_uid), Some(&alice_fpr),
             true,
             &[
                 (1, "[✓] ")
             ])
            .expect("valid");

        test("certification-network.pgp",
             &dave_fpr,
             &["--certification-network"], "authenticate", &[],
             120, Some(&alice_uid), Some(&alice_fpr),
             false,
             &[
                 (1, "[ ] ")
             ])
            .expect("valid");

        test("certification-network.pgp",
             &dave_fpr,
             &["--certification-network"], "authenticate", &[],
             60, Some(&alice_uid), Some(&alice_fpr),
             true,
             &[
                 (1, "[✓] ")
             ])
            .expect("valid");

        test("certification-network.pgp",
             &carol_fpr,
             &[], "authenticate", &[],
             60, Some(&alice_uid), Some(&alice_fpr),
             false,
             &[
                 (1, "[ ] ")
             ])
            .expect("valid");

        test("certification-network.pgp",
             &carol_fpr,
             &["--certification-network"], "authenticate", &[],
             60, Some(&alice_uid), Some(&alice_fpr),
             true,
             &[
                 (1, "[✓] ")
             ])
            .expect("valid");

        test("certification-network.pgp",
             &carol_fpr,
             &["--certification-network"], "authenticate", &[],
             60, Some(&bob_uid), Some(&bob_fpr),
             true,
             &[
                 (1, "[✓] ")
             ])
            .expect("valid");

        Ok(())
    }

    #[test]
    #[allow(unused)]
    fn path_certification_network() -> Result<()> {
        let alice_fpr: Fingerprint =
            "B2B371214EF71AFD16E42C62D81360B4C0489225"
           .parse().expect("valid fingerprint");
        let alice_uid
            = UserID::from("<alice@example.org>");
        // Certified by: 9A1AE937B5CB8BC46048AB63023CC01973ED9DF3

        let bob_fpr: Fingerprint =
            "A68DF00EB82F9C49C27CC7723C5F5BBE6B790C05"
           .parse().expect("valid fingerprint");
        let bob_uid
            = UserID::from("<bob@example.org>");
        // Certified by: B2B371214EF71AFD16E42C62D81360B4C0489225

        let carol_fpr: Fingerprint =
            "AB9EF1C89631519842ED559697557DD147D99C97"
           .parse().expect("valid fingerprint");
        let carol_uid
            = UserID::from("<carol@example.org>");
        // Certified by: A68DF00EB82F9C49C27CC7723C5F5BBE6B790C05

        let dave_fpr: Fingerprint =
            "9A1AE937B5CB8BC46048AB63023CC01973ED9DF3"
           .parse().expect("valid fingerprint");
        let dave_uid
            = UserID::from("<dave@example.org>");
        // Certified by: AB9EF1C89631519842ED559697557DD147D99C97

        // Alice does not make Bob a trusted introducer.  So without
        // --certificate-network, she can only authenticate Bob, but
        // not Carol or Dave.
        test("certification-network.pgp",
             &alice_fpr,
             &[], "path",
             &[ &alice_fpr.to_string(),
                &bob_fpr.to_string(),
                &bob_uid.to_string()
             ],
             120, None, None,
             true,
             &[
                 (1, "[✓] ")
             ])
            .expect("valid");

        test("certification-network.pgp",
             &alice_fpr,
             &[], "path",
             &[ &alice_fpr.to_string(),
                &bob_fpr.to_string(),
                &carol_fpr.to_string(),
                &carol_uid.to_string()
             ],
             120, None, None,
             false,
             &[
                 (1, "[ ] ")
             ])
            .expect("valid");

        test("certification-network.pgp",
             &alice_fpr,
             &[], "path",
             &[ &alice_fpr.to_string(),
                &bob_fpr.to_string(),
                &carol_fpr.to_string(),
                &dave_fpr.to_string(),
                &dave_uid.to_string()
             ],
             120, None, None,
             false,
             &[
                 (1, "[ ] ")
             ])
            .expect("valid");

        // With --certification-network, she can authenticate them all.
        test("certification-network.pgp",
             &alice_fpr,
             &["--certification-network"], "path",
             &[ &alice_fpr.to_string(),
                &bob_fpr.to_string(),
                &bob_uid.to_string()
             ],
             120, None, None,
             true,
             &[
                 (1, "[✓] ")
             ])
            .expect("valid");

        test("certification-network.pgp",
             &alice_fpr,
             &["--certification-network"], "path",
             &[ &alice_fpr.to_string(),
                &bob_fpr.to_string(),
                &carol_fpr.to_string(),
                &carol_uid.to_string()
             ],
             120, None, None,
             true,
             &[
                 (1, "[✓] ")
             ])
            .expect("valid");

        test("certification-network.pgp",
             &alice_fpr,
             &["--certification-network"], "path",
             &[ &alice_fpr.to_string(),
                &bob_fpr.to_string(),
                &carol_fpr.to_string(),
                &dave_fpr.to_string(),
                &dave_uid.to_string()
             ],
             120, None, None,
             true,
             &[
                 (1, "[✓] ")
             ])
            .expect("valid");

        // But invalid paths should stay invalid.
        test("certification-network.pgp",
             &alice_fpr,
             &["--certification-network"], "path",
             &[ &alice_fpr.to_string(),
                &carol_fpr.to_string(),
                &carol_uid.to_string()
             ],
             120, None, None,
             false,
             &[
                 (1, "[ ] ")
             ])
            .expect("valid");


        // dave authenticates alice for 60 of 120.
        test("certification-network.pgp",
             &dave_fpr,
             &[], "path",
             &[ &dave_fpr.to_string(),
                &alice_fpr.to_string(),
                &alice_uid.to_string()
             ],
             120, None, None,
             false,
             &[
                 (1, "[ ] ")
             ])
            .expect("valid");

        test("certification-network.pgp",
             &dave_fpr,
             &[], "path",
             &[ &dave_fpr.to_string(),
                &alice_fpr.to_string(),
                &alice_uid.to_string()
             ],
             60, None, None,
             true,
             &[
                 (1, "[✓] ")
             ])
            .expect("valid");

        test("certification-network.pgp",
             &dave_fpr,
             &["--certification-network"], "path",
             &[ &dave_fpr.to_string(),
                &alice_fpr.to_string(),
                &alice_uid.to_string()
             ],
             120, None, None,
             false,
             &[
                 (1, "[ ] ")
             ])
            .expect("valid");

        test("certification-network.pgp",
             &dave_fpr,
             &["--certification-network"], "path",
             &[ &dave_fpr.to_string(),
                &alice_fpr.to_string(),
                &alice_uid.to_string()
             ],
             60, None, None,
             true,
             &[
                 (1, "[✓] ")
             ])
            .expect("valid");

        test("certification-network.pgp",
             &carol_fpr,
             &[], "path",
             &[ &carol_fpr.to_string(),
                &dave_fpr.to_string(),
                &alice_fpr.to_string(),
                &alice_uid.to_string()
             ],
             60, None, None,
             false,
             &[
                 (1, "[ ] ")
             ])
            .expect("valid");

        test("certification-network.pgp",
             &carol_fpr,
             &["--certification-network"], "path",
             &[ &carol_fpr.to_string(),
                &dave_fpr.to_string(),
                &alice_fpr.to_string(),
                &alice_uid.to_string()
             ],
             60, None, None,
             true,
             &[
                 (1, "[✓] ")
             ])
            .expect("valid");

        test("certification-network.pgp",
             &carol_fpr,
             &["--certification-network"], "path",
             &[ &carol_fpr.to_string(),
                &dave_fpr.to_string(),
                &alice_fpr.to_string(),
                &bob_fpr.to_string(),
                &bob_uid.to_string()
             ],
             60, None, None,
             true,
             &[
                 (1, "[✓] ")
             ])
            .expect("valid");

        Ok(())
    }

    #[test]
    #[allow(unused)]
    fn gossip_certification_network() -> Result<()> {
        let alice_fpr: Fingerprint =
            "B2B371214EF71AFD16E42C62D81360B4C0489225"
           .parse().expect("valid fingerprint");
        let alice_uid
            = UserID::from("<alice@example.org>");
        // Certified by: 9A1AE937B5CB8BC46048AB63023CC01973ED9DF3

        let bob_fpr: Fingerprint =
            "A68DF00EB82F9C49C27CC7723C5F5BBE6B790C05"
           .parse().expect("valid fingerprint");
        let bob_uid
            = UserID::from("<bob@example.org>");
        // Certified by: B2B371214EF71AFD16E42C62D81360B4C0489225

        let carol_fpr: Fingerprint =
            "AB9EF1C89631519842ED559697557DD147D99C97"
           .parse().expect("valid fingerprint");
        let carol_uid
            = UserID::from("<carol@example.org>");
        // Certified by: A68DF00EB82F9C49C27CC7723C5F5BBE6B790C05

        let dave_fpr: Fingerprint =
            "9A1AE937B5CB8BC46048AB63023CC01973ED9DF3"
           .parse().expect("valid fingerprint");
        let dave_uid
            = UserID::from("<dave@example.org>");
        // Certified by: AB9EF1C89631519842ED559697557DD147D99C97

        // Alice certified Bob.  We should print the path, but it
        // should be unauthenticated (this is gossip).
        test("certification-network.pgp",
             None,
             &["--gossip"], "authenticate", &[ ],
             120, Some(&bob_uid), Some(&bob_fpr),
             true,
             &[
                 (2, format!("[ ] {}", bob_fpr.to_hex()))
             ])
            .expect("valid");

        // Make sure we don't authenticate when we specify a root
        // (which is ignored when --gossip is provided).
        test("certification-network.pgp",
             &alice_fpr,
             &["--gossip"], "authenticate", &[ ],
             120, Some(&bob_uid), Some(&bob_fpr),
             true,
             &[
                 (2, format!("[ ] {}", bob_fpr.to_hex()))
             ])
            .expect("valid");

        Ok(())
    }

    #[test]
    #[allow(unused)]
    fn target_cert_expired() -> Result<()> {
        let alice_fpr: Fingerprint =
            "1FA62523FB7C06E71EEFB82BB5159F3FC3EB3AC9"
           .parse().expect("valid fingerprint");
        let alice_uid
            = UserID::from("<alice@example.org>");

        let bob_fpr: Fingerprint =
            "B166B31AE5F95600B3F7184FE74C6CE62821686F"
           .parse().expect("valid fingerprint");
        let bob_uid
            = UserID::from("<bob@example.org>");
        // Certified by: 1FA62523FB7C06E71EEFB82BB5159F3FC3EB3AC9

        let carol_fpr: Fingerprint =
            "81CD118AC5BD9156DC113772626222D76ACDFFCF"
           .parse().expect("valid fingerprint");
        let carol_uid
            = UserID::from("<carol@example.org>");
        // Certified by: B166B31AE5F95600B3F7184FE74C6CE62821686F

        // $ date '+%s' -d 20200202
        // 1580598000
        let t1 = std::time::UNIX_EPOCH + time::Duration::new(1580598000, 0);
        // $ date '+%s' -d 20200302
        // 1583103600
        let t2 = std::time::UNIX_EPOCH + time::Duration::new(1583103600, 0);
        // $ date '+%s' -d 20200402
        // 1585778400
        let t3 = std::time::UNIX_EPOCH + time::Duration::new(1585778400, 0);

        // At t1, Alice certifies Bob (amount = 60).
        // At t2, Bob's certificate expires.

        // Bob's certificate is not yet expired.
        test("cert-expired.pgp",
             Some(&alice_fpr),
             &[], "authenticate", &[ "--time", "20200214" ],
             60, Some(&bob_uid), Some(&bob_fpr),
             true,
             &[
                 (1, "[✓] ")
             ])
            .expect("valid");

        test("cert-expired.pgp",
             None,
             &[], "path",
             &[ "--time", "20200214",
                   &alice_fpr.to_string(),
                   &bob_fpr.to_string(),
                   &bob_uid.to_string() ],
             60, None, None,
             true,
             &[
                 (1, "[✓] ")
             ])
            .expect("valid");

        // Bob's certificate is expired.
        test("cert-expired.pgp",
             Some(&alice_fpr),
             &[], "authenticate", &[ "--time", "20200216" ],
             60, Some(&bob_uid), Some(&bob_fpr),
             false,
             &[
                 (0, "[✓] ")
             ])
            .expect("valid");

        test("cert-expired.pgp",
             None,
             &[], "path",
             &[ "--time", "20200216",
                   &alice_fpr.to_string(),
                   &bob_fpr.to_string(),
                   &bob_uid.to_string() ],
             60, None, None,
             false,
             &[
                 (1, "[ ] ")
             ])
            .expect("valid");

        Ok(())
    }

    #[test]
    #[allow(unused)]
    fn target_cert_hard_revoked() -> Result<()> {
        let alice_fpr: Fingerprint =
            "219AAB661C8AAF4526DBC31AA751A7A0532863BA"
           .parse().expect("valid fingerprint");
        let alice_uid
            = UserID::from("<alice@example.org>");

        let bob_fpr: Fingerprint =
            "90E02BFB03FAA04714D1D3D87543157EF3B12BE9"
           .parse().expect("valid fingerprint");
        let bob_uid
            = UserID::from("<bob@example.org>");
        // Certified by: 219AAB661C8AAF4526DBC31AA751A7A0532863BA
        // Certified by: 219AAB661C8AAF4526DBC31AA751A7A0532863BA

        let carol_fpr: Fingerprint =
            "BF680710128E6BCCB2268154569F5F6BFB95C544"
           .parse().expect("valid fingerprint");
        let carol_uid
            = UserID::from("<carol@example.org>");
        // Certified by: 219AAB661C8AAF4526DBC31AA751A7A0532863BA

        let dave_fpr: Fingerprint =
            "46945292F8F643F0573AF71183F9C1A4759A16D6"
           .parse().expect("valid fingerprint");
        let dave_uid
            = UserID::from("<dave@example.org>");
        // Certified by: 90E02BFB03FAA04714D1D3D87543157EF3B12BE9
        // Certified by: BF680710128E6BCCB2268154569F5F6BFB95C544
        // Certified by: 90E02BFB03FAA04714D1D3D87543157EF3B12BE9

        // Bob's certificate is hard revoked.
        test("cert-revoked-hard.pgp",
             Some(&alice_fpr),
             &[], "authenticate", &[],
             60, Some(&bob_uid), Some(&bob_fpr),
             false,
             &[
                 (1, "[ ] ")
             ])
            .expect("valid");

        test("cert-revoked-hard.pgp",
             None,
             &[], "path",
             &[
                   &alice_fpr.to_string(),
                   &bob_fpr.to_string(),
                   &bob_uid.to_string() ],
             60, None, None,
             false,
             &[
                 (1, "[ ] "),
                 (1, "revoked "),
             ])
            .expect("valid");

        Ok(())
    }


    #[test]
    #[allow(unused)]
    fn target_cert_soft_revoked() -> Result<()> {
        let alice_fpr: Fingerprint =
            "66037F98B444BBAFDFE98E871738DFAB86878262"
           .parse().expect("valid fingerprint");
        let alice_uid
            = UserID::from("<alice@example.org>");

        let bob_fpr: Fingerprint =
            "4CD8737F76C2B897C4F058DBF28C47540FA2C3B3"
           .parse().expect("valid fingerprint");
        let bob_uid
            = UserID::from("<bob@example.org>");
        // Certified by: 66037F98B444BBAFDFE98E871738DFAB86878262

        let carol_fpr: Fingerprint =
            "AB4E3F8EE8BBD3459754D75ACE570F9B8C7DC75D"
           .parse().expect("valid fingerprint");
        let carol_uid
            = UserID::from("<carol@example.org>");
        // Certified by: 66037F98B444BBAFDFE98E871738DFAB86878262

        let dave_fpr: Fingerprint =
            "DF6A440ED9DE723B0EBC7F50E24FBB1B9FADC999"
           .parse().expect("valid fingerprint");
        let dave_uid
            = UserID::from("<dave@example.org>");
        // Certified by: 4CD8737F76C2B897C4F058DBF28C47540FA2C3B3
        // Certified by: AB4E3F8EE8BBD3459754D75ACE570F9B8C7DC75D
        // Certified by: 4CD8737F76C2B897C4F058DBF28C47540FA2C3B3

        // Bob's certificate is soft revoked on 20200301.  If the
        // reference time is before that, we should be able to
        // authenticate Bob.  After that and we should fail to do so.
        test("cert-revoked-soft.pgp",
             Some(&alice_fpr),
             &["--time", "20200228"], "authenticate", &[],
             60, Some(&bob_uid), Some(&bob_fpr),
             true,
             &[
                 (1, "[✓] ")
             ])
            .expect("valid");

        test("cert-revoked-soft.pgp",
             None,
             &["--time", "20200228"], "path",
             &[
                   &alice_fpr.to_string(),
                   &bob_fpr.to_string(),
                   &bob_uid.to_string() ],
             60, None, None,
             true,
             &[
                 (1, "[✓] ")
             ])
            .expect("valid");

        test("cert-revoked-soft.pgp",
             Some(&alice_fpr),
             &["--time", "20200302"], "authenticate", &[],
             60, Some(&bob_uid), Some(&bob_fpr),
             false,
             &[
                 (1, "[ ] ")
             ])
            .expect("valid");

        test("cert-revoked-soft.pgp",
             None,
             &["--time", "20200302"], "path",
             &[
                   &alice_fpr.to_string(),
                   &bob_fpr.to_string(),
                   &bob_uid.to_string() ],
             60, None, None,
             false,
             &[
                 (1, "[ ] "),
                 (1, "revoked "),
             ])
            .expect("valid");

        Ok(())
    }

    #[test]
    #[allow(unused)]
    fn target_userid_revoked() -> Result<()> {
        let alice_fpr: Fingerprint =
            "01672BB67E4B4047E5A4EC0A731CEA092C465FC8"
           .parse().expect("valid fingerprint");
        let alice_uid
            = UserID::from("<alice@example.org>");

        let bob_fpr: Fingerprint =
            "EA479A77CD074458EAFE56B4861BF42FF490C581"
           .parse().expect("valid fingerprint");
        let bob_uid
            = UserID::from("<bob@example.org>");
        // Certified by: 01672BB67E4B4047E5A4EC0A731CEA092C465FC8
        // Certified by: 01672BB67E4B4047E5A4EC0A731CEA092C465FC8

        let carol_fpr: Fingerprint =
            "212873BB9C4CC49F8E5A6FEA78BC5397470BA7F0"
           .parse().expect("valid fingerprint");
        let carol_uid
            = UserID::from("<carol@example.org>");
        // Certified by: EA479A77CD074458EAFE56B4861BF42FF490C581
        // Certified by: EA479A77CD074458EAFE56B4861BF42FF490C581

        // Bob's User ID is soft revoked on 20200301.  If the
        // reference time is before that, we should be able to
        // authenticate Bob.  After that and we should fail to do so.
        test("userid-revoked.pgp",
             Some(&alice_fpr),
             &["--time", "20200228"], "authenticate", &[],
             60, Some(&bob_uid), Some(&bob_fpr),
             true,
             &[
                 (1, "[✓] ")
             ])
            .expect("valid");

        test("userid-revoked.pgp",
             None,
             &["--time", "20200228"], "path",
             &[
                   &alice_fpr.to_string(),
                   &bob_fpr.to_string(),
                   &bob_uid.to_string() ],
             60, None, None,
             true,
             &[
                 (1, "[✓] ")
             ])
            .expect("valid");

        test("userid-revoked.pgp",
             Some(&alice_fpr),
             &["--time", "20200302"], "authenticate", &[],
             60, Some(&bob_uid), Some(&bob_fpr),
             false,
             &[
                 (1, "[ ] ")
             ])
            .expect("valid");

        test("userid-revoked.pgp",
             None,
             &["--time", "20200302"], "path",
             &[
                   &alice_fpr.to_string(),
                   &bob_fpr.to_string(),
                   &bob_uid.to_string() ],
             60, None, None,
             false,
             &[
                 (1, "[ ] "),
                 (1, "revoked "),
             ])
            .expect("valid");

        Ok(())
    }
}
